from typing import Annotated

import typer

from typerdrive.cache.attach import get_manager, attach_cache
from typerdrive.cache.exceptions import CacheError
from typerdrive.cache.manager import CacheManager
from typerdrive.exceptions import handle_errors
from typerdrive.format import terminal_message


@handle_errors("Failed to clear cache", handle_exc_class=CacheError)
@attach_cache()
def clear(
    ctx: typer.Context,
    path: Annotated[
        str | None,
        typer.Option(help="Clear only the entry matching this path. If not provided, clear the entire cache"),
    ] = None,
):
    manager: CacheManager = get_manager(ctx)
    if path:
        manager.clear_path(path)
        terminal_message(f"Cleared entry at cache target {str(path)}")
    else:
        typer.confirm("Are you sure you want to clear the entire cache?", abort=True)
        count = manager.clear_all()
        terminal_message(f"Cleared all {count} files from cache")


def add_clear(cli: typer.Typer):
    cli.command()(clear)


@handle_errors("Failed to show cache", handle_exc_class=CacheError)
@attach_cache(show=True)
def show(ctx: typer.Context):  # pyright: ignore[reportUnusedParameter]
    pass


def add_show(cli: typer.Typer):
    cli.command()(show)


def add_cache_subcommand(cli: typer.Typer):
    cache_cli = typer.Typer(help="Manage cache for the app")

    for cmd in [add_clear, add_show]:
        cmd(cache_cli)

    cli.add_typer(cache_cli, name="cache")



import json
from pathlib import Path

import typer

from typerdrive.constants import ExitCode
from typerdrive.settings.commands import add_bind, add_reset, add_settings_subcommand, add_show, add_unset, add_update

from tests.helpers import match_help, match_output
from tests.settings.models import DefaultSettingsModel, RequiredFieldsModel


class TestBind:
    def test_bind__updates_settings_with_cli_args(self, fake_settings_path: Path):
        cli = typer.Typer()

        add_bind(cli, RequiredFieldsModel)

        expected_pattern = [
            "name.*hutt",
            "planet.*nal hutta",
            "is-humanoid.*False",
            "alignment.*neutral",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            "--name=hutt",
            "--planet=nal hutta",
            "--no-is-humanoid",
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        assert fake_settings_path.exists()
        data = json.loads(fake_settings_path.read_text())
        assert data == dict(
            name="hutt",
            planet="nal hutta",
            is_humanoid=False,
            alignment="neutral",
        )

    def test_bind__raises_exception_if_final_config_is_not_valid(self):
        cli = typer.Typer()

        add_bind(cli, RequiredFieldsModel)

        match_output(
            cli,
            "--name=hutt",
            "--planet=nal hutta",
            "--alignment=invalid-alignment",
            exit_code=ExitCode.GENERAL_ERROR,
            expected_pattern=[
                "Failed to bind settings",
                "Final settings are invalid",
                "RequiredFieldsModel",
                "invalid-alignment is an invalid alignment",
            ],
            prog_name="test",
        )

    def test_bind__help_text(self):
        cli = typer.Typer()

        add_bind(cli, RequiredFieldsModel)

        expected_pattern = [
            "Options",
            r"--name TEXT \[default: None\] \[required\]",
            r"--planet TEXT \[default: None\] \[required\]",
            r"--is-humanoid --no-is-humanoid  \[default: is-humanoid\]",
            r"--alignment TEXT \[default: neutral\]",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )


class TestUpdate:
    def test_update__updates_settings_with_cli_args(self, fake_settings_path: Path):
        cli = typer.Typer()

        add_update(cli, DefaultSettingsModel)

        expected_pattern = [
            "name.*hutt",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            "--name=hutt",
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        assert fake_settings_path.exists()
        data = json.loads(fake_settings_path.read_text())
        assert data == dict(
            name="hutt",
            planet="tatooine",
            is_humanoid=True,
            alignment="neutral",
        )

    def test_update__does_not_fail_with_invalid_settings(self, fake_settings_path: Path):
        cli = typer.Typer()

        add_update(cli, RequiredFieldsModel)

        expected_pattern = [
            "name.*hutt",
            "is-humanoid.*True",
            "alignment.*neutral",
            "Settings are invalid:",
            "planet.*Field required",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            "--name=hutt",
            expected_pattern=expected_pattern,
            prog_name="test",
        )

    def test_update__help_text(self):
        cli = typer.Typer()

        add_update(cli, RequiredFieldsModel)

        expected_pattern = [
            "Options",
            r"--name TEXT \[default: None\]",
            r"--planet TEXT \[default: None\]",
            r"--is-humanoid",  # NOTE: This doesn't match typer docs. It should show a default
            r"--alignment TEXT \[default: None\]",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )


class TestUnset:
    def test_unset__unsets_fields_with_cli_args(self, fake_settings_path: Path):
        fake_settings_path.write_text(
            json.dumps(
                dict(
                    name="pyke",
                    planet="oba diah",
                    is_humanoid=True,
                    alignment="evil",
                ),
            ),
        )

        cli = typer.Typer()

        add_unset(cli, RequiredFieldsModel)

        expected_pattern = [
            "name.*pyke",
            "planet.*oba diah",
            "is-humanoid.*True",
            "alignment.*neutral",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            "--alignment",
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        assert fake_settings_path.exists()
        data = json.loads(fake_settings_path.read_text())
        assert data == dict(
            name="pyke",
            planet="oba diah",
            is_humanoid=True,
            alignment="neutral",
        )

    def test_unset__does_not_fail_with_invalid_settings(self, fake_settings_path: Path):
        fake_settings_path.write_text(
            json.dumps(
                dict(
                    name="pyke",
                    planet="oba diah",
                    is_humanoid=True,
                    alignment="evil",
                ),
            ),
        )

        cli = typer.Typer()

        add_unset(cli, RequiredFieldsModel)

        expected_pattern = [
            "is-humanoid.*True",
            "alignment.*evil",
            "Settings are invalid:",
            "name.*Field required",
            "planet.*Field required",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            "--name",
            "--planet",
            expected_pattern=expected_pattern,
            prog_name="test",
        )

    def test_unset__help_text(self):
        cli = typer.Typer()

        add_unset(cli, RequiredFieldsModel)

        expected_pattern = [
            "Options",
            r"--name (?! \[default)",
            r"--planet (?! \[default)",
            r"--is-humanoid (?! \[default)",
            r"--alignment (?! \[default)",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )


class TestShow:
    def test_show__shows_settings(self):
        cli = typer.Typer()

        add_show(cli, DefaultSettingsModel)

        expected_pattern = [
            "name.*jawa",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

    def test_show__help_text(self):
        cli = typer.Typer()

        add_show(cli, RequiredFieldsModel)

        expected_pattern = [
            r"name",
            r"planet",
            r"is-humanoid",
            r"alignment",
        ]
        match_help(
            cli,
            unwanted_pattern=expected_pattern,
            prog_name="test",
        )


class TestReset:
    def test_reset__resets_all_fields_to_defaults(self, fake_settings_path: Path):
        fake_settings_path.write_text(
            json.dumps(
                dict(
                    name="pyke",
                    planet="oba diah",
                    is_humanoid=True,
                    alignment="evil",
                ),
            ),
        )

        cli = typer.Typer()

        add_reset(cli, RequiredFieldsModel)

        expected_pattern = [
            "is-humanoid.*True",
            "alignment.*neutral",
            "Settings are invalid:",
            "name.*Field required",
            "planet.*Field required",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
            input="y",
        )

        assert fake_settings_path.exists()
        data = json.loads(fake_settings_path.read_text())
        assert data == dict(
            is_humanoid=True,
            alignment="neutral",
        )

    def test_reset__help_text(self):
        cli = typer.Typer()

        add_reset(cli, RequiredFieldsModel)

        expected_pattern = [
            r"name",
            r"planet",
            r"is-humanoid",
            r"alignment",
        ]
        match_help(
            cli,
            unwanted_pattern=expected_pattern,
            prog_name="test",
        )


class TestSubcommand:
    def test_add_settings_subcommand(self):
        cli = typer.Typer()

        add_settings_subcommand(cli, RequiredFieldsModel)

        expected_pattern = [
            r"settings.*Manage settings for the app",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        expected_pattern = [
            "bind",
            "update",
            "unset",
            "reset",
            "show",
        ]
        match_output(
            cli,
            "settings",
            "--help",
            exit_code=0,
            expected_pattern=expected_pattern,
            prog_name="test",
        )



import io
import inspect
from pathlib import Path
import re
import sys
import tempfile
import textwrap
from importlib import import_module
from typing import Any
from collections.abc import Callable
from dataclasses import dataclass

import snick
from rich import box
from rich.console import Console, Group, RenderableType
from rich.panel import Panel
from rich.prompt import Confirm
from rich.markdown import Markdown
from rich.rule import Rule

from typerdrive.env import tweak_env


@dataclass
class Decomposed:
    module: str
    name: str
    docstring: str
    source: str


@dataclass
class Captured:
    error: Exception | None = None
    stdout: str | None = None
    stderr: str | None = None
    exit_code: int | str | None = None
    settings: str | None = None


BlankLine = Rule(characters=" ")


def pseudo_clear(console: Console):
    for _ in range(console.size.height):
        console.print(BlankLine)


def fake_input(text: str):
    sys.stdin.write(text)
    sys.stdin.seek(0)


def get_demo_functions(module_name: str) -> list[Callable[..., None]]:
    demo_functions: list[Callable[..., None]] = []
    module = import_module(f"typerdrive_demo.{module_name}")
    for _, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) and obj.__name__.startswith("demo"):
            demo_functions.append(obj)
    return sorted(demo_functions, key=lambda f: f.__name__)


def decompose(func: Callable[..., Any]) -> Decomposed:
    """
    This is really hacky.

    Maybe improve this sometime.
    """
    module = func.__module__.split(".")[-1]
    name = func.__name__

    if func.__doc__ is None:
        raise RuntimeError("Can't demo a function with no docstring!")
    docstring = textwrap.dedent(func.__doc__).strip()

    source_lines = inspect.getsourcelines(func)[0]
    first_quotes = False
    code_start_index = 0
    for i, line in enumerate(source_lines):
        if line.strip() == '"""':
            if not first_quotes:
                first_quotes = True
            else:
                code_start_index = i
                break
    if code_start_index == 0:
        raise RuntimeError("Failed to strip function declaration and docstring!")
    source_lines = [re.sub(r"\s+# pyright.*", "", sl) for sl in source_lines]
    source_lines = [re.sub(r"\s+# type.*", "", sl) for sl in source_lines]
    source = textwrap.dedent("".join(source_lines[code_start_index + 1 :]))

    return Decomposed(module=module, name=name, docstring=docstring, source=source)


def capture(demo: Callable[..., None]) -> Captured:
    demo_name = demo.__name__

    cap = Captured()

    stdout_buffer = io.StringIO()
    original_stdout = sys.stdout
    sys.stdout = stdout_buffer

    stderr_buffer = io.StringIO()
    original_stderr = sys.stderr
    sys.stderr = stderr_buffer

    stdin_buffer = io.StringIO()
    original_stdin = sys.stdin
    sys.stdin = stdin_buffer

    original_argv = sys.argv
    sys.argv = [demo_name]

    with tempfile.TemporaryDirectory() as fake_home:
        fake_settings_dir = Path(fake_home) / ".local/share" / demo_name
        fake_settings_dir.mkdir(parents=True)
        fake_settings_path = fake_settings_dir / "settings.json"
        with tweak_env(COLUMNS="80", HOME=fake_home):
            try:
                demo()
            except SystemExit as exited:
                cap.exit_code = exited.code
            except Exception as exc:
                cap.error = exc
            finally:
                if fake_settings_path.exists():
                    cap.settings = fake_settings_path.read_text()
                sys.stdout = original_stdout
                sys.stderr = original_stderr
                sys.stdin = original_stdin
                sys.argv = original_argv

    stdout_dump = stdout_buffer.getvalue()
    if stdout_dump:
        cap.stdout = stdout_dump

    stderr_dump = stderr_buffer.getvalue()
    if stderr_dump:
        cap.stderr = stderr_dump

    return cap


def run_demo(demo: Callable[..., None], console: Console, override_label: str | None = None) -> bool:
    pseudo_clear(console)

    decomposed = decompose(demo)
    cap: Captured = capture(demo)

    parts: list[RenderableType] = [
        Markdown(decomposed.docstring),
        BlankLine,
        BlankLine,
        Panel(
            Markdown(
                "\n".join(
                    [
                        "```python",
                        decomposed.source,
                        "```",
                    ]
                )
            ),
            title=f"Here is the source code for [yellow]{decomposed.name}()[/yellow]",
            title_align="left",
            padding=1,
            expand=False,
            box=box.SIMPLE,
        ),
    ]

    if cap.stdout:
        parts.extend(
            [
                BlankLine,
                BlankLine,
                Panel(
                    Markdown(
                        snick.conjoin("```text", cap.stdout, "```"),
                    ),
                    title=f"Here is the stdout captured from [yellow]{decomposed.name}()[/yellow]",
                    title_align="left",
                    padding=1,
                    expand=False,
                    box=box.SIMPLE,
                ),
            ]
        )

    if cap.stderr:
        parts.extend(
            [
                BlankLine,
                BlankLine,
                Panel(
                    Markdown(
                        snick.conjoin("```text", cap.stderr, "```"),
                    ),
                    title=f"Here is the stderr captured from [yellow]{decomposed.name}()[/yellow]",
                    title_align="left",
                    padding=1,
                    expand=False,
                    box=box.SIMPLE,
                ),
            ]
        )

    if cap.error:
        parts.extend(
            [
                BlankLine,
                BlankLine,
                Panel(
                    f"[red]{cap.error.__class__.__name__}[/red]: [yellow]{str(cap.error)}[/yellow]",
                    title=f"Here is the uncaught exception from [yellow]{decomposed.name}()[/yellow]",
                    title_align="left",
                    padding=1,
                    expand=False,
                    box=box.SIMPLE,
                ),
            ]
        )

    if cap.exit_code is not None:
        parts.extend(
            [
                BlankLine,
                BlankLine,
                Panel(
                    Markdown(
                        snick.conjoin("```text", str(cap.exit_code), "```"),
                    ),
                    title=f"Here is the exit code from [yellow]{decomposed.name}()[/yellow]",
                    title_align="left",
                    padding=1,
                    expand=False,
                    box=box.SIMPLE,
                ),
            ]
        )

    if cap.settings is not None:
        parts.extend(
            [
                BlankLine,
                BlankLine,
                Panel(
                    Markdown(
                        snick.conjoin("```json", cap.settings, "```"),
                    ),
                    title=f"Here are the final contents of the settings file from [yellow]{decomposed.name}()[/yellow]",
                    title_align="left",
                    padding=1,
                    expand=False,
                    box=box.SIMPLE,
                ),
            ]
        )

    label = override_label if override_label else f"{decomposed.module}()"
    console.print(
        Panel(
            Group(*parts),
            padding=1,
            title=f"Showing [yellow]{decomposed.name}()[/yellow] for [green]{label}[/green]",
            title_align="left",
            subtitle="[blue]https://github.com/dusktreader/typerdrive[/blue]",
            subtitle_align="left",
        ),
    )
    console.print(BlankLine)
    console.print(BlankLine)
    further: bool = Confirm.ask("Would you like to continue?", default=True)
    return further



from typing import Annotated

import typer
from pydantic import BaseModel, AfterValidator
from snick import unwrap

from typerdrive.settings.attach import attach_settings


def valid_alignment(value: str) -> str:
    if value not in ["good", "neutral", "evil"]:
        raise ValueError(f"{value} is an invalid alignment")
    return value


class SettingsModel(BaseModel):
    name: str = "jawa"
    planet: str = "tatooine"
    is_humanoid: bool = True
    alignment: Annotated[str, AfterValidator(valid_alignment)] = "neutral"


cli = typer.Typer()


@cli.command()
@attach_settings(SettingsModel)
def report(ctx: typer.Context, cfg: SettingsModel):
    print(
        unwrap(
            f"""
            Look at this {cfg.alignment} {cfg.name} from {cfg.planet}
            {'walking' if cfg.is_humanoid else 'slithering'} by.
            """
        )
    )


if __name__ == "__main__":
    cli()



from collections.abc import Generator
from pathlib import Path

import pytest

from typerdrive.env import tweak_env


@pytest.fixture
def fake_settings_path(tmp_path: Path) -> Generator[Path, None, None]:
    fake_settings_dir = tmp_path / ".local/share/test"
    fake_settings_dir.mkdir(parents=True)
    settings_path = fake_settings_dir / "settings.json"

    with tweak_env(HOME=str(tmp_path)):
        yield settings_path



import typer

from typerdrive.cache.attach import attach_cache
from typerdrive.cache.manager import CacheManager
from typerdrive.cache.exceptions import CacheError


cli = typer.Typer()


@cli.command()
@attach_cache()
def report(ctx: typer.Context, manager: CacheManager):
    path = "jawa/ewok.txt"
    text: str
    try:
        text = manager.load_text(path)
    except CacheError:
        print("Cache miss! Created new text")
        text = "Never will you find a more wretched hive of scum and villainy."
    else:
        print(f"Cache hit! Loaded text from cache target {path}")

    print(f"Text: {text}")
    manager.store_text(text, path)
    print(f"Stored text at cache target {path}")


if __name__ == "__main__":
    cli()



from datetime import datetime
from typing import cast

from pydantic import BaseModel

from typerdrive.client.base import TyperdriveClient


def demo_1__typerdrive_client__using_a_query_param_model():
    """
    This function demonstrates how the `_x()` methods of the `TyperdriveClient`
    can use a pydantic model to provide query parameters to the request. In this
    case, the `get_x()` method takes a `param_obj` keyword argument that provides
    all the query parameters that are used in the request. You may not use the
    `params` keyword argument used by normal `httx.Client` instances if you use
    the `param_obj` parameter.
    """

    class QueryParams(BaseModel):
        search: str


    class People(BaseModel, extra="ignore"):
        name: str
        birth_year: str
        gender: str

    class ResponseModel(BaseModel, extra="ignore"):
        results: list[People]


    client = TyperdriveClient(base_url="https://swapi.py4e.com/api/")
    param_obj = QueryParams(search="skywalker")
    response = cast(
        ResponseModel,
        client.get_x("people/", param_obj=param_obj, response_model=ResponseModel),
    )
    print("People ->", ", ".join(r.name for r in response.results))


def demo_2__typerdrive_client__using_a_request_body_model():
    """
    This function demonstrates how the `_x()` methods of the `TyperdriveClient`
    can use a pydantic model to provide a request body. In this case, the
    `post_x()` method takes a `body_obj` keyword argument that provides the
    complete body for the request. You may not use other keyword arguments such
    as `data`, `json`, or `content` if you use the `body_obj` parameter.
    """

    class PeopleData(BaseModel):
        birth_year: str
        gender: str
        species: str


    class Body(BaseModel):
        name: str
        data: PeopleData


    class ResponseModel(BaseModel, extra="ignore"):
        id: str
        name: str
        createdAt: datetime
        data: PeopleData


    client = TyperdriveClient(base_url="https://api.restful-api.dev/")
    body_obj = Body(name="Mara Jade Skywalker", data=PeopleData(birth_year="17 BBY", gender="Female", species="Human"))
    response = cast(
        ResponseModel,
        client.post_x("objects", body_obj=body_obj, response_model=ResponseModel),
    )
    print("New Person ->", response)


def demo_3__typerdrive_client__expecting_specific_status():
    """
    This function demonstrates how the `_x()` methods of the `TyperdriveClient`
    can specify the exact status code that they want to receive. If that status
    code is not returned, they will raise a `ClientError`.
    """

    client = TyperdriveClient(base_url="https://swapi.py4e.com/api/")
    client.get_x("people/1/", expected_status=203)


def demo_4__typerdrive_client__with_no_expected_response():
    """
    This function demonstrates how the `_x()` methods of the `TyperdriveClient`
    can disregard the response and only return the status code. This is useful
    for any endpoints that do may not return a body, and you are only concerned
    that the request was successful.
    """
    client = TyperdriveClient(base_url="https://swapi.py4e.com/api/")
    response = client.get_x("people/1/", expect_response=False)
    print("Status Code ->", response)


def demo_5__typerdrive_client__with_no_response_model():
    """
    This function demonstrates how the `_x()` methods of the `TyperdriveClient`
    can return a plain dictionary from the JSON body of the response if no
    `response_model` is provided.
    """
    client = TyperdriveClient(base_url="https://swapi.py4e.com/api/")
    response = client.get_x("people/1/")
    print("Raw Data ->", response)



import typer
from pydantic import BaseModel
from pytest_mock import MockerFixture

from typerdrive.constants import ExitCode
from typerdrive.exceptions import TyperdriveError, handle_errors
from typerdrive.cache.manager import CacheManager
from typerdrive.cache.attach import attach_cache
from typerdrive.settings.attach import attach_settings, get_settings

from tests.helpers import check_output


class TestHandleErrors:
    def test_handle_errors__no_errors(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?")
        def _():
            print("We're all fine here now, thank you. How are you?")

        check_output(cli, expected_substring="thank you", unwanted_substring="What happened?", prog_name="test")

    def test_handle_errors__no_conflict_with_attach_cache(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?")
        @attach_cache()
        def _(ctx: typer.Context):
            manager = ctx.obj.cache_manager
            assert isinstance(manager, CacheManager)
            assert manager.app_name == "test"
            print("We're all fine here now, thank you. How are you?")

        check_output(cli, expected_substring="thank you", unwanted_substring="What happened?", prog_name="test")

    def test_handle_errors__no_conflict_with_attach_settings(self):
        class SettingsModel(BaseModel):
            name: str = "jawa"

        cli: typer.Typer = typer.Typer()

        @cli.command()
        @attach_settings(SettingsModel)
        def _(ctx: typer.Context):
            settings: SettingsModel = get_settings(ctx, SettingsModel)
            assert settings.name == "jawa"
            print("We're all fine here now, thank you. How are you?")

        check_output(cli, expected_substring="thank you", unwanted_substring="What happened?", prog_name="test")

    def test_handle_errors__handles_typerdrive_error_by_default(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?")
        def _():
            raise TyperdriveError("Boring conversation anyway")

        check_output(
            cli,
            expected_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            unwanted_substring="Traceback",
            exit_code=TyperdriveError.exit_code,
            prog_name="test",
        )

    def test_handle_errors__handles_errors_matching_handle_exc_class(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?", handle_exc_class=RuntimeError)
        def han():  # pyright: ignore[reportUnusedFunction]
            raise RuntimeError("Boring conversation anyway")

        @cli.command()
        @handle_errors("Let me see your identification.", handle_exc_class=RuntimeError)
        def ben():  # pyright: ignore[reportUnusedFunction]
            raise ValueError("These aren't the droids you're looking for")

        check_output(
            cli,
            "han",
            expected_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            unwanted_substring="Traceback",
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )

        check_output(
            cli,
            "ben",
            unwanted_substring=[
                "Let me see your identification",
                "These aren't the droids you're looking for",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            exception_type=ValueError,
            exception_pattern="These aren't the droids you're looking for",
            prog_name="test",
        )

    def test_handle_errors__handles_all_errors_in_handle_exc_class_group(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?", handle_exc_class=(RuntimeError, ValueError))
        def han():  # pyright: ignore[reportUnusedFunction]
            raise RuntimeError("Boring conversation anyway")

        @cli.command()
        @handle_errors("Let me see your identification.", handle_exc_class=(RuntimeError, ValueError))
        def ben():  # pyright: ignore[reportUnusedFunction]
            raise ValueError("These aren't the droids you're looking for")

        check_output(
            cli,
            "han",
            expected_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            unwanted_substring="Traceback",
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )

        check_output(
            cli,
            "ben",
            expected_substring=[
                "Let me see your identification",
                "These aren't the droids you're looking for",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )

    def test_handle_errors__ignores_errors_matching_ignore_exc_class(self):
        cli: typer.Typer = typer.Typer()

        class BlasterRuntimeError(RuntimeError):
            pass

        class LightSaberRuntimeError(RuntimeError):
            pass

        @cli.command()
        @handle_errors(
            "What happened?",
            handle_exc_class=RuntimeError,
            ignore_exc_class=BlasterRuntimeError,
        )
        def han():  # pyright: ignore[reportUnusedFunction]
            raise BlasterRuntimeError("Boring conversation anyway")

        @cli.command()
        @handle_errors(
            "Let me see your identification.",
            handle_exc_class=RuntimeError,
            ignore_exc_class=BlasterRuntimeError,
        )
        def ben():  # pyright: ignore[reportUnusedFunction]
            raise LightSaberRuntimeError("These aren't the droids you're looking for")

        check_output(
            cli,
            "han",
            unwanted_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            exception_type=BlasterRuntimeError,
            exception_pattern="Boring conversation anyway",
            prog_name="test",
        )

        check_output(
            cli,
            "ben",
            expected_substring=[
                "Let me see your identification",
                "These aren't the droids you're looking for",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )

    def test_handle_errors__ignores_all_errors_in_ignore_exc_class_group(self):
        cli: typer.Typer = typer.Typer()

        class BlasterRuntimeError(RuntimeError):
            pass

        class LightSaberRuntimeError(RuntimeError):
            pass

        @cli.command()
        @handle_errors(
            "What happened?",
            handle_exc_class=RuntimeError,
            ignore_exc_class=(BlasterRuntimeError, LightSaberRuntimeError),
        )
        def han():  # pyright: ignore[reportUnusedFunction]
            raise BlasterRuntimeError("Boring conversation anyway")

        @cli.command()
        @handle_errors(
            "Let me see your identification.",
            handle_exc_class=RuntimeError,
            ignore_exc_class=(BlasterRuntimeError, LightSaberRuntimeError),
        )
        def ben():  # pyright: ignore[reportUnusedFunction]
            raise LightSaberRuntimeError("These aren't the droids you're looking for")

        check_output(
            cli,
            "han",
            unwanted_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            exception_type=BlasterRuntimeError,
            exception_pattern="Boring conversation anyway",
            prog_name="test",
        )

        check_output(
            cli,
            "ben",
            unwanted_substring=[
                "Let me see your identification",
                "These aren't the droids you're looking for",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            exception_type=LightSaberRuntimeError,
            exception_pattern="These aren't the droids you're looking for",
            prog_name="test",
        )

    def test_handle_errors__do_except(self):
        cli: typer.Typer = typer.Typer()

        garbage_chute: list[str] = []

        @cli.command()
        @handle_errors("What happened?", do_except=lambda params: garbage_chute.append(params.base_message))
        def _():
            raise TyperdriveError("Boring conversation anyway")

        check_output(
            cli,
            expected_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )
        assert "What happened?" in garbage_chute

    def test_handle_errors__do_else(self):
        cli: typer.Typer = typer.Typer()

        garbage_chute: list[str] = []

        @cli.command()
        @handle_errors("What happened?", do_else=lambda: garbage_chute.append("some rescue"))
        def _():
            print("We're all fine here now, thank you. How are you?")

        check_output(
            cli,
            expected_substring="We're all fine here now, thank you. How are you?",
            unwanted_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            prog_name="test",
        )
        assert "some rescue" in garbage_chute

    def test_handle_errors__do_finally(self):
        cli: typer.Typer = typer.Typer()

        garbage_chute: list[str] = []

        @cli.command()
        @handle_errors("What happened?", do_finally=lambda: garbage_chute.append("I don't care what you smell"))
        def _():
            raise TyperdriveError("Boring conversation anyway")

        check_output(
            cli,
            expected_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )
        assert "I don't care what you smell" in garbage_chute

    def test_handle_errors__raises_runtime_error_if_format_fails(self, mocker: MockerFixture):
        mocker.patch("typerdrive.exceptions.reformat_exception", side_effect=ValueError("BOOM!"))

        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?")
        def _():
            raise TyperdriveError("Boring conversation anyway")

        check_output(
            cli,
            unwanted_substring=[
                "What happened?",
                "Boring conversation anyway",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            exception_type=RuntimeError,
            exception_pattern="Failed while formatting message.*BOOM!",
            prog_name="test",
        )

    def test_handle_errors__uses_subject_and_footer_from_typerdrive_error_if_provided(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @handle_errors("What happened?")
        def _():
            raise TyperdriveError(
                "Boring conversation anyway",
                subject="slight weapons malfunction",
                footer="WE'RE GONNA HAVE COMPANY!",
            )

        check_output(
            cli,
            expected_substring=[
                "slight weapons malfunction",
                "Boring conversation anyway",
                "WE'RE GONNA HAVE COMPANY!",
            ],
            exit_code=ExitCode.GENERAL_ERROR,
            prog_name="test",
        )



from typing import Any

import typer
from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from typer_repyt.build_command import DecDef, build_command, OptDef
from typer_repyt.constants import Sentinel

from typerdrive.constants import Validation
from typerdrive.exceptions import handle_errors
from typerdrive.settings.attach import get_manager, attach_settings
from typerdrive.settings.exceptions import SettingsError
from typerdrive.settings.manager import SettingsManager


def bind(ctx: typer.Context):
    __manager: SettingsManager = get_manager(ctx)
    excluded_locals = ["__manager", "ctx"]
    settings_values = {k: v for (k, v) in locals().items() if k not in excluded_locals}
    __manager.update(**settings_values)


def add_bind(cli: typer.Typer, settings_model: type[BaseModel]):
    opt_defs: list[OptDef] = []
    for name, field_info in settings_model.model_fields.items():
        default = field_info.default if field_info.default is not PydanticUndefined else Sentinel.NOT_GIVEN
        param_type: type[Any] = SettingsError.enforce_defined(
            field_info.annotation, "Option type may not be `None`"
        )  # TODO: Figure out if this can even be triggered
        opt_defs.append(
            OptDef(
                name=name,
                param_type=param_type,
                default=default,
                help=field_info.description,
                show_default=True,
            )
        )
    build_command(
        cli,
        bind,
        *opt_defs,
        decorators=[
            DecDef(
                dec_func=handle_errors,
                dec_args=["Failed to bind settings"],
                dec_kwargs=dict(
                    handle_exc_class=SettingsError,
                    unwrap_message=False,
                    debug=True,
                ),
                is_simple=False,
            ),
            DecDef(
                dec_func=attach_settings,
                dec_args=[settings_model],
                dec_kwargs=dict(validation=Validation.AFTER, persist=True, show=True),
                is_simple=False,
            ),
        ],
        include_context=True,
    )


def update(ctx: typer.Context):
    __manager: SettingsManager = get_manager(ctx)
    excluded_locals = ["__manager", "ctx"]
    settings_values = {k: v for (k, v) in locals().items() if k not in excluded_locals and v is not None}
    __manager.update(**settings_values)


def add_update(cli: typer.Typer, settings_model: type[BaseModel]):
    opt_defs: list[OptDef] = []
    for name, field_info in settings_model.model_fields.items():
        param_type: type[Any] = SettingsError.enforce_defined(field_info.annotation, "Option type may not be `None`")
        default: None | bool = None
        if param_type is bool:
            default = field_info.default
        opt_defs.append(
            OptDef(
                name=name,
                param_type=param_type | None,
                default=default,
                help=field_info.description,
                show_default=True,
            )
        )
    build_command(
        cli,
        update,
        *opt_defs,
        decorators=[
            DecDef(
                dec_func=handle_errors,
                dec_args=["Failed to update settings"],
                dec_kwargs=dict(
                    handle_exc_class=SettingsError,
                    unwrap_message=False,
                    debug=True,
                ),
                is_simple=False,
            ),
            DecDef(
                dec_func=attach_settings,
                dec_args=[settings_model],
                dec_kwargs=dict(validation=Validation.NONE, persist=True, show=True),
                is_simple=False,
            ),
        ],
        include_context=True,
    )


# TODO: Don't forget to add docstrings
def unset(ctx: typer.Context):
    __manager: SettingsManager = get_manager(ctx)
    excluded_locals = ["__manager", "ctx"]
    settings_values = {k: v for (k, v) in locals().items() if k not in excluded_locals and v}
    __manager.unset(*settings_values.keys())


def add_unset(cli: typer.Typer, settings_model: type[BaseModel]):
    opt_defs: list[OptDef] = []
    for name, field_info in settings_model.model_fields.items():
        opt_defs.append(
            OptDef(
                name=name,
                param_type=bool,
                default=False,
                help=field_info.description,
                show_default=True,
                override_name=name,
            )
        )
    build_command(
        cli,
        unset,
        *opt_defs,
        decorators=[
            DecDef(
                dec_func=handle_errors,
                dec_args=["Failed to unset settings"],
                dec_kwargs=dict(
                    handle_exc_class=SettingsError,
                    unwrap_message=False,
                    debug=True,
                ),
                is_simple=False,
            ),
            DecDef(
                dec_func=attach_settings,
                dec_args=[settings_model],
                dec_kwargs=dict(validation=Validation.NONE, persist=True, show=True),
                is_simple=False,
            ),
        ],
        include_context=True,
    )


def show(ctx: typer.Context):  # pyright: ignore[reportUnusedParameter]
    pass


def add_show(cli: typer.Typer, settings_model: type[BaseModel]):
    build_command(
        cli,
        show,
        decorators=[
            DecDef(
                dec_func=handle_errors,
                dec_args=["Failed to show settings"],
                dec_kwargs=dict(
                    handle_exc_class=SettingsError,
                    unwrap_message=False,
                    debug=True,
                ),
                is_simple=False,
            ),
            DecDef(
                dec_func=attach_settings,
                dec_args=[settings_model],
                dec_kwargs=dict(validation=Validation.NONE, persist=False, show=True),
                is_simple=False,
            ),
        ],
        include_context=True,
    )


def reset(ctx: typer.Context):
    typer.confirm("Are you sure you want to reset your settings?", abort=True)
    __manager: SettingsManager = get_manager(ctx)
    __manager.reset()


def add_reset(cli: typer.Typer, settings_model: type[BaseModel]):
    build_command(
        cli,
        reset,
        decorators=[
            DecDef(
                dec_func=handle_errors,
                dec_args=["Failed to reset settings"],
                dec_kwargs=dict(
                    handle_exc_class=SettingsError,
                    unwrap_message=False,
                    debug=True,
                ),
                is_simple=False,
            ),
            DecDef(
                dec_func=attach_settings,
                dec_args=[settings_model],
                dec_kwargs=dict(validation=Validation.NONE, persist=True, show=True),
                is_simple=False,
            ),
        ],
        include_context=True,
    )


def add_settings_subcommand(cli: typer.Typer, settings_model: type[BaseModel]):
    settings_cli = typer.Typer(help="Manage settings for the app")

    for cmd in [add_bind, add_update, add_unset, add_reset, add_show]:
        cmd(settings_cli, settings_model)

    cli.add_typer(settings_cli, name="settings")



import json
from pathlib import Path
from typing import Annotated
from collections.abc import Generator

import pytest
import snick
from pydantic import AfterValidator, BaseModel, ValidationError
from pytest_mock import MockerFixture

from typerdrive.env import tweak_env
from typerdrive.settings.exceptions import SettingsInitError, SettingsSaveError, SettingsUpdateError
from typerdrive.settings.manager import SettingsManager, get_settings_path


def valid_alignment(value: str) -> str:
    if value not in ["good", "neutral", "evil"]:
        raise ValueError(f"{value} is an invalid alignment")
    return value


class DefaultSettingsModel(BaseModel):
    name: str = "jawa"
    planet: str = "tatooine"
    is_humanoid: bool = True
    alignment: Annotated[str, AfterValidator(valid_alignment)] = "neutral"


class RequiredFieldsModel(BaseModel):
    name: str
    planet: str
    is_humanoid: bool = True
    alignment: Annotated[str, AfterValidator(valid_alignment)] = "neutral"


@pytest.fixture
def fake_settings_path(tmp_path: Path) -> Generator[Path, None, None]:
    fake_settings_dir = tmp_path / ".local/share/test"
    fake_settings_dir.mkdir(parents=True)
    settings_path = fake_settings_dir / "settings.json"

    with tweak_env(HOME=str(tmp_path)):
        yield settings_path


class TestSettingsManager:
    def test_init__loads_default_model(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        assert manager.app_name == "test"
        assert manager.settings_model == DefaultSettingsModel
        assert manager.settings_path == get_settings_path("test")
        assert manager.invalid_warnings == {}

        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "jawa"
        assert manager.settings_instance.planet == "tatooine"
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"

    def test_init__loads_default_model_from_disk(self, fake_settings_path: Path):
        settings_path = fake_settings_path
        settings_path.write_text(
            json.dumps(
                dict(name="hutt", planet="nal hutta", is_humanoid=False, alignment="evil"),
            )
        )

        manager = SettingsManager("test", DefaultSettingsModel)

        assert manager.app_name == "test"
        assert manager.settings_model == DefaultSettingsModel
        assert manager.settings_path == settings_path
        assert manager.invalid_warnings == {}

        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "hutt"
        assert manager.settings_instance.planet == "nal hutta"
        assert not manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "evil"

    def test_init__loads_invalid_model_with_warnings(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        assert manager.app_name == "test"
        assert manager.settings_model == RequiredFieldsModel
        assert manager.settings_path == get_settings_path("test")
        assert manager.invalid_warnings == dict(
            name="Field required",
            planet="Field required",
        )

        assert isinstance(manager.settings_instance, RequiredFieldsModel)
        assert not hasattr(manager.settings_instance, "name")
        assert not hasattr(manager.settings_instance, "planet")
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"

    def test_init__loads_invalid_model_from_disk(self, fake_settings_path: Path):
        settings_path = fake_settings_path
        settings_path.write_text(json.dumps(dict(name="hutt")))

        manager = SettingsManager("test", RequiredFieldsModel)

        assert manager.app_name == "test"
        assert manager.settings_model == RequiredFieldsModel
        assert manager.settings_path == settings_path
        assert manager.invalid_warnings == dict(
            planet="Field required",
        )

        assert isinstance(manager.settings_instance, RequiredFieldsModel)
        assert manager.settings_instance.name == "hutt"
        assert not hasattr(manager.settings_instance, "planet")
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"

    def test_init__raises_error_on_failure(self, mocker: MockerFixture):
        mocker.patch.object(DefaultSettingsModel, "__init__", side_effect=RuntimeError("BOOM"))
        with pytest.raises(SettingsInitError, match="Failed to initialize settings"):
            SettingsManager("test", DefaultSettingsModel)

    def test_set_warnings__unpacks_validation_error(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        err = ValidationError.from_exception_data(
            "BOOM",
            [
                {
                    "type": "missing",
                    "loc": ("name",),
                    "input": dict(),
                },
                {
                    "type": "missing",
                    "loc": ("planet",),
                    "input": dict(),
                },
            ],
        )
        manager.set_warnings(err)

        assert manager.invalid_warnings == dict(
            name="Field required",
            planet="Field required",
        )

    def test_update__updates_valid_to_valid_settings(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        manager.update(planet="nevarro", alignment="good")
        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "jawa"
        assert manager.settings_instance.planet == "nevarro"
        assert manager.settings_instance.alignment == "good"
        assert manager.settings_instance.is_humanoid
        assert manager.invalid_warnings == {}

    def test_update__updates_invalid_to_valid_settings(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        assert len(manager.invalid_warnings) > 0
        manager.update(name="pyke", planet="oba diah")
        assert isinstance(manager.settings_instance, RequiredFieldsModel)
        assert manager.settings_instance.name == "pyke"
        assert manager.settings_instance.planet == "oba diah"
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"
        assert manager.invalid_warnings == {}

    def test_update__updates_valid_to_invalid_settings(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        manager.update(planet="nevarro", alignment="negative")
        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "jawa"
        assert manager.settings_instance.planet == "nevarro"
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "negative"
        assert manager.invalid_warnings == dict(
            alignment="Value error, negative is an invalid alignment",
        )

    def test_update__raises_error_on_failure(self, mocker: MockerFixture):
        manager = SettingsManager("test", DefaultSettingsModel)
        mocker.patch.object(manager.settings_model, "__init__", side_effect=RuntimeError("BOOM"))
        with pytest.raises(SettingsUpdateError, match="Failed to update settings"):
            manager.update()

    def test_unset__resets_field_to_default(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        manager.update(name="pyke", planet="oba diah", alignment="evil")
        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "pyke"
        assert manager.settings_instance.planet == "oba diah"
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "evil"
        assert manager.invalid_warnings == {}
        manager.unset("name", "alignment")
        assert manager.settings_instance.name == "jawa"
        assert manager.settings_instance.planet == "oba diah"
        assert manager.settings_instance.alignment == "neutral"

    def test_unset__adds_warnings_if_field_has_no_default(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        assert isinstance(manager.settings_instance, RequiredFieldsModel)
        manager.update(name="jawa", planet="tatooine", alignment="neutral")
        assert manager.invalid_warnings == {}
        manager.unset("planet")
        assert manager.settings_instance.name == "jawa"
        assert not hasattr(manager.settings_instance, "planet")
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"
        assert manager.invalid_warnings == dict(
            planet="Field required",
        )

    def test_reset__returns_all_fields_to_default(self):
        manager = SettingsManager("test", DefaultSettingsModel)
        manager.update(name="hutt", planet="nal hutta", is_humanoid=False, alignment="evil")
        assert isinstance(manager.settings_instance, DefaultSettingsModel)
        assert manager.settings_instance.name == "hutt"
        assert manager.settings_instance.planet == "nal hutta"
        assert not manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "evil"
        assert manager.invalid_warnings == {}
        manager.reset()
        assert manager.settings_instance.name == "jawa"
        assert manager.settings_instance.planet == "tatooine"
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"

    def test_reset__adds_warnings_for_fields_with_no_defaults(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        manager.update(name="hutt", planet="nal hutta", is_humanoid=False, alignment="evil")
        assert isinstance(manager.settings_instance, RequiredFieldsModel)
        assert manager.settings_instance.name == "hutt"
        assert manager.settings_instance.planet == "nal hutta"
        assert not manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "evil"
        assert manager.invalid_warnings == {}
        manager.reset()
        assert not hasattr(manager.settings_instance, "name")
        assert not hasattr(manager.settings_instance, "planet")
        assert manager.settings_instance.is_humanoid
        assert manager.settings_instance.alignment == "neutral"
        assert manager.invalid_warnings == dict(
            name="Field required",
            planet="Field required",
        )

    def test_pretty__provides_string_reflecting_settings_instance(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        manager.update(name="pyke", alignment="negative")
        computed = manager.pretty()
        print(computed)
        expected = snick.dedent(
            """
            [bold]       name[/bold] -> pyke
            [bold]is-humanoid[/bold] -> True
            [bold]  alignment[/bold] -> [red]negative[/red]

            [red]Settings are invalid:[/red]
            [bold]     planet[/bold] -> Field required
            [bold]  alignment[/bold] -> Value error, negative is an invalid alignment
            """
        ).strip()
        print(expected)
        assert expected == computed

    def test_pretty__without_style(self):
        manager = SettingsManager("test", RequiredFieldsModel)
        manager.update(name="pyke", alignment="negative")
        computed = manager.pretty(with_style=False)
        print(computed)
        expected = (
            "       "
            + snick.dedent(
                """
                   name -> pyke
            is-humanoid -> True
              alignment -> negative

            Settings are invalid:
                 planet -> Field required
              alignment -> Value error, negative is an invalid alignment
            """
            ).strip()
        )
        print(expected)
        assert expected == computed

    def test_save__writes_settings_to_disk(self, fake_settings_path: Path):
        manager = SettingsManager("test", DefaultSettingsModel)
        manager.save()

        assert json.loads(fake_settings_path.read_text()) == dict(
            name="jawa", planet="tatooine", is_humanoid=True, alignment="neutral"
        )

    def test_save__raises_error_on_failure(self, mocker: MockerFixture):
        manager = SettingsManager("test", DefaultSettingsModel)
        mocker.patch.object(manager, "settings_path").write_text.side_effect = RuntimeError("BOOM")
        with pytest.raises(SettingsSaveError, match="Failed to save settings"):
            manager.save()



import json

import httpx
import pydantic
import pytest
import respx

from typerdrive.client.base import TyperdriveClient
from typerdrive.client.exceptions import ClientError


class QueryParams(pydantic.BaseModel):
    anger: bool
    fear: bool
    aggression: bool


class RequestBody(pydantic.BaseModel):
    anger: bool
    fear: bool
    aggression: bool


class ResponseModel(pydantic.BaseModel):
    speed: str
    difficulty: str
    attractiveness: str


class TestRequestX:

    def test_request_x__basic(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.request_x("GET", "the-dark-side")

        assert mock_route.called
        assert isinstance(response, dict)
        assert response == dict(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


    def test_request_x__with_param_obj(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                ),
            )
        )
        client.request_x("GET", "the-dark-side", param_obj=QueryParams(anger=True, fear=True, aggression=True))

        assert mock_route.called
        assert dict(mock_route.calls.last.request.url.params) == dict(anger="true", fear="true", aggression="true")


    def test_request_x__raises_exception_when_params_provided_with_param_obj(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock()
        with pytest.raises(ClientError, match="'params' not allowed"):
            client.request_x(
                "GET",
                "the-dark-side",
                params=dict(anger=True),
                param_obj=QueryParams(anger=True, fear=True, aggression=True)
            )

        assert not mock_route.called


    def test_request_x__raises_exception_for_invalid_param_obj(self, respx_mock: respx.MockRouter):

        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock()
        with pytest.raises(ClientError, match="Param data could not be deserialized"):
            client.request_x("GET", "the-dark-side", param_obj="not a pydantic model instance")  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]

        assert not mock_route.called


    def test_request_x__with_body_obj(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.post("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                ),
            )
        )
        client.request_x("POST", "the-dark-side", body_obj=RequestBody(anger=True, fear=True, aggression=True))

        assert mock_route.called
        assert mock_route.calls.last.request.headers["Content-Type"] == "application/json"
        assert json.loads(mock_route.calls.last.request.content.decode("utf-8")) == dict(
            anger=True,
            fear=True,
            aggression=True,
        )


    def test_request_x__raises_exception_when_body_provided_with_body_obj(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.post("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                ),
            )
        )
        with pytest.raises(ClientError, match="'data', 'json' and 'content' not allowed"):
            client.request_x(
                "POST",
                "the-dark-side",
                json=dict(anger=True),
                body_obj=RequestBody(anger=True, fear=True, aggression=True)
            )

        assert not mock_route.called


    def test_request_x__raises_exception_for_invalid_body_model(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.post("https://the.force.io/the-dark-side").mock()
        with pytest.raises(ClientError, match="Request body data could not be deserialized"):
            client.request_x(
                "POST",
                "the-dark-side",
                body_obj="not a pydantic model instance",  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]
            )

        assert not mock_route.called


    def test_request_x__raises_exception_when_httpx_request_fails(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        respx_mock.get("https://the.force.io/the-dark-side").mock(side_effect=httpx.RequestError("Boom!"))
        with pytest.raises(ClientError, match="Communication with the API failed"):
            client.request_x("GET", "the-dark-side")


    def test_request_x__raises_exception_when_status_code_does_not_match_expected(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(status_code=202)
        )

        with pytest.raises(ClientError, match="Got an unexpected status code: Expected 200, got 202"):
            client.request_x("GET", "the-dark-side", expected_status=200)

        assert mock_route.called


    def test_request_x__just_returns_status_code_if_not_expect_response(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(status_code=204)
        )

        response = client.request_x("GET", "the-dark-side", expect_response=False)

        assert mock_route.called
        assert response == 204


    def test_request_x__raises_exception_if_response_does_not_unpack(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(status_code=200, content="NOT SERIALIZABLE"),
        )

        with pytest.raises(ClientError, match="Failed to unpack JSON"):
            client.request_x("GET", "the-dark-side")

        assert mock_route.called


    def test_request_x__returns_dict_if_response_model_is_not_provided(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.request_x("GET", "the-dark-side")

        assert mock_route.called
        assert isinstance(response, dict)
        assert response == dict(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


    def test_request_x__deserializes_to_response_model_if_provided(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.request_x("GET", "the-dark-side", response_model=ResponseModel)

        assert mock_route.called
        assert isinstance(response, ResponseModel)
        assert response == ResponseModel(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


    def test_request_x__raises_exception_when_body_cannot_be_deserialized_to_response_model(self, respx_mock: respx.MockRouter):

        class OtherModel(pydantic.BaseModel):
            planet: str

        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        with pytest.raises(ClientError, match="Unexpected data in response"):
            client.request_x("GET", "the-dark-side", response_model=OtherModel)

        assert mock_route.called


class TestGetX:

    def test_get_x(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.get("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.get_x(
            "the-dark-side",
            param_obj=QueryParams(anger=True, fear=True, aggression=True),
            response_model=ResponseModel,
        )

        assert mock_route.called
        assert isinstance(response, ResponseModel)
        assert response == ResponseModel(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


class TestPostX:

    def test_post_x(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.post("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.post_x(
            "the-dark-side",
            body_obj=RequestBody(anger=True, fear=True, aggression=True),
            response_model=ResponseModel,
        )

        assert mock_route.called
        assert isinstance(response, ResponseModel)
        assert response == ResponseModel(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


class TestPutX:

    def test_put_x(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.put("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.put_x(
            "the-dark-side",
            body_obj=RequestBody(anger=True, fear=True, aggression=True),
            response_model=ResponseModel,
        )

        assert mock_route.called
        assert isinstance(response, ResponseModel)
        assert response == ResponseModel(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )


class TestPatchX:

    def test_put_x(self, respx_mock: respx.MockRouter):
        client = TyperdriveClient(base_url="https://the.force.io")

        mock_route = respx_mock.patch("https://the.force.io/the-dark-side").mock(
            return_value=httpx.Response(
                status_code=200,
                json=dict(
                    speed="quicker",
                    difficulty="easier",
                    attractiveness="seductive"
                )
            )
        )

        response = client.patch_x(
            "the-dark-side",
            body_obj=RequestBody(anger=True, fear=True, aggression=True),
            response_model=ResponseModel,
        )

        assert mock_route.called
        assert isinstance(response, ResponseModel)
        assert response == ResponseModel(
            speed="quicker",
            difficulty="easier",
            attractiveness="seductive"
        )



from typing import Any, Literal

import snick
from rich.console import Console, RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text


def terminal_message(
    message: str | RenderableType,
    subject: str | None = None,
    subject_align: Literal["left", "right", "center"] = "left",
    color: str = "green",
    footer: str | None = None,
    footer_align: Literal["left", "right", "center"] = "left",
    indent: bool = True,
    markdown: bool = False,
    error: bool = False,
):
    panel_kwargs: dict[str, Any] = dict(padding=1, title_align=subject_align, subtitle_align=footer_align)
    if subject is not None:
        panel_kwargs["title"] = f"[{color}]{subject}"
    if footer is not None:
        panel_kwargs["subtitle"] = f"[dim italic]{footer}[/dim italic]"
    if isinstance(message, str):
        message = snick.dedent(message)
        if indent:
            message = snick.indent(message, prefix="  ")
        if markdown:
            message = Markdown(message)
    console = Console(stderr=error)
    console.print()
    console.print(Panel(message, **panel_kwargs))
    console.print()


def simple_message(message: str, indent: bool = False, markdown: bool = False, error: bool = False):
    text: str = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    content: str | Markdown = text
    if markdown:
        content = Markdown(text)
    console = Console(stderr=error)
    console.print()
    console.print(content)
    console.print()


def strip_rich_style(text: str | Text) -> str:
    if isinstance(text, str):
        text = Text.from_markup(text)
    return text.plain



"""
This set of demos shows the use of the various settings-related commands.
"""

import json
from pathlib import Path

import typer
from pydantic import BaseModel
from typerdrive.settings.commands import add_settings_subcommand
from typerdrive_demo.helpers import fake_input


def demo_1__bind__basic():
    """
    This function demonstrates the use of the `bind` settings command.
    This command allows you to set the settings values for your app as
    defined in the settings model. It will allow you to override
    defaults. After updating the settings model, the updated values are
    written to disk where they can be retrieved for future commands.
    After the `bind` command is complete, it shows the updated settings.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "bind", "--name=jawa", "--planet=tatooine"])


def demo_2__bind__invalid_settings():
    """
    This function demonstrates how the `bind` settings command
    requires all settings that do not have a default to be provided
    a value through the command line. If any of these required values
    are not included, the CLI app will show an error message.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "bind", "--name=jawa"])


def demo_3__bind__help():
    """
    This function demonstrates the help text that's provided with
    the `bind` command.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "bind", "--help"])


def demo_4__update():
    """
    This function demonstrates the use of the `update` settings command.
    This command allows you to set _some_ of the settings values for your
    app as defined in the settings model. Unlike the `bind` command,
    `update` allows the settings to have missing values that result in an
    invalid state. Any settings options that are not provided are ignored.
    After the `update` command is complete, it shows the updated settings
    including any settings values that are still invalid.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "update", "--planet=tatooine"])


def demo_5__update__help():
    """
    This function demonstrates the help text that's provided with
    the `update` command.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "update", "--help"])


def demo_6__unset():
    """
    This function demonstrates the use of the `unset` settings command.
    This command allows you to return settings values to their defaults.
    If a selected settings value does not have a default, it will just
    be removed from the settings. The `unset` command allows the final
    settings to be in an invalid state. After the `unset` command is
    complete, it shows the updated settings including any settings
    values that are still invalid.

    Note:
        To demonstrate this working, we need to hack the settings file
        before the command is invoked.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    settings_path = Path.home() / ".local/share/demo_6__unset/settings.json"
    settings_path.write_text(
        json.dumps(
            dict(
                name="hutt",
                planet="nal hutta",
                is_humanoid=False,
                alignment="evil",
            ),
        ),
    )

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "unset", "--name", "--alignment"])


def demo_7__unset__help():
    """
    This function demonstrates the help text that's provided with
    the `unset` command.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "unset", "--help"])


def demo_8__show():
    """
    This function demonstrates the `show` command. It simply shows
    the current settings.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)
    cli(["settings", "show"])


def demo_9__reset():
    """
    This function demonstrates the `reset` command. It clears all the
    settings. All settings values with defaults are reset to their
    defaults. All settings values without defaults are just removed.
    After the `reset` command is complete, it shows the updated settings
    including any settings values that are now invalid.

    Note:
        To demonstrate this working, we need to hack the settings file
        before the command is invoked. We also need to add a forced
        confirmation because this command requires you to confirm before
        it proceeds.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str
        is_humanoid: bool = True
        alignment: str = "neutral"

    settings_path = Path.home() / ".local/share/demo_9__reset/settings.json"
    settings_path.write_text(
        json.dumps(
            dict(
                name="hutt",
                planet="nal hutta",
                is_humanoid=False,
                alignment="evil",
            ),
        ),
    )

    cli = typer.Typer()
    add_settings_subcommand(cli, ExampleSettings)

    fake_input("y")
    cli(["settings", "reset"])



"""
This set of demos shows the use of the various settings-related commands.
"""

import typer
from typerdrive.cache.commands import add_cache_subcommand
from typerdrive.cache.manager import CacheManager

from typerdrive_demo.helpers import fake_input


def demo_1__clear__specific_path():
    """
    This function demonstrates the use of the `clear` cache command
    targeting a specific item of cached data identified by its path.
    """

    manager = CacheManager("demo_1__clear__specific_path")
    manager.store_text("Utinni!", "jawa.txt")
    manager.store_bytes(b"Yub Nub!", "ewok")
    manager.store_json(
        dict(
            full_name="Jabba Desilijic Tiure",
            age=604,
            quote="beeska chata wnow kong bantha poodoo!",
        ),
        "hutt/jabba.json",
    )

    cli = typer.Typer()
    add_cache_subcommand(cli)
    cli(["cache", "clear", "--path=jawa.txt"])


def demo_2__clear__full_cache():
    """
    This function demonstrates the use of the `clear` cache command
    clearing out the entire cache. Note that the command requires
    confirmation before proceeding with the deletion.
    """

    manager = CacheManager("demo_2__clear__full_cache")
    manager.store_text("Utinni!", "jawa.txt")
    manager.store_bytes(b"Yub Nub!", "ewok")
    manager.store_json(
        dict(
            full_name="Jabba Desilijic Tiure",
            age=604,
            quote="beeska chata wnow kong bantha poodoo!",
        ),
        "hutt/jabba.json",
    )

    cli = typer.Typer()
    add_cache_subcommand(cli)

    fake_input("y")
    cli(["cache", "clear"])


def demo_3__show():
    """
    This function demonstrates the use of the `show` cache command.
    This command shows the full directory structure of the cache
    including the size of the files within it.
    """

    manager = CacheManager("demo_3__show")
    manager.store_text("Utinni!", "jawa.txt")
    manager.store_bytes(b"Yub Nub!", "ewok")
    manager.store_json(
        dict(
            full_name="Jabba Desilijic Tiure",
            age=604,
            quote="beeska chata wnow kong bantha poodoo!",
        ),
        "hutt/jabba.json",
    )

    cli = typer.Typer()
    add_cache_subcommand(cli)

    cli(["cache", "show"])



import json
from pathlib import Path
from typing import cast

import pytest
import typer
from pydantic import BaseModel
from typer.testing import CliRunner

from typerdrive.constants import Validation
from typerdrive.settings.attach import attach_settings, get_manager, get_settings
from typerdrive.settings.exceptions import SettingsError
from typerdrive.settings.manager import SettingsManager

from tests.helpers import match_output, match_help
from tests.settings.models import DefaultSettingsModel, RequiredFieldsModel


class TestAttachSettings:
    def test_attach_settings__adds_settings_to_context(self, runner: CliRunner):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            settings = get_settings(ctx, DefaultSettingsModel)
            assert settings.name == "jawa"
            assert settings.planet == "tatooine"
            assert settings.is_humanoid
            assert settings.alignment == "neutral"

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 0

    def test_attach_settings__fails_on_invalid_settings_by_default(self, runner: CliRunner):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(RequiredFieldsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            pass

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 1

    def test_attach_settings__does_not_fail_on_invalid_settings_with_no_validation(self, runner: CliRunner):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(RequiredFieldsModel, validation=Validation.NONE)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            settings = get_settings(ctx, RequiredFieldsModel)
            assert not hasattr(settings, "name")
            assert not hasattr(settings, "planet")
            assert settings.is_humanoid
            assert settings.alignment == "neutral"

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 0

    def test_attach_settings__fails_before_with_before_validation(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(RequiredFieldsModel, validation=Validation.BEFORE)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            print("in function body")

        expected_pattern = [
            "in function body",
        ]
        match_output(
            cli,
            exit_code=1,
            unwanted_pattern=expected_pattern,
            exception_type=SettingsError,
            exception_pattern="Initial settings are invalid",
            prog_name="test",
        )

    def test_attach_settings__fails_after_with_after_validation(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(RequiredFieldsModel, validation=Validation.AFTER)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            print("in function body")

        expected_pattern = [
            "in function body",
        ]
        match_output(
            cli,
            exit_code=1,
            expected_pattern=expected_pattern,
            exception_type=SettingsError,
            exception_pattern="Final settings are invalid",
            prog_name="test",
        )

    @pytest.mark.parametrize(
        "model,exception_pattern",
        [
            (RequiredFieldsModel, "Initial settings are invalid"),
            (DefaultSettingsModel, "Final settings are invalid"),
        ],
    )
    def test_attach_settings__fails_before_and_after_with_both_validation(
        self,
        model: type[BaseModel],
        exception_pattern: str,
    ):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(model, validation=Validation.BOTH)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            print("in function body")
            settings = get_settings(ctx, model)
            setattr(settings, "alignment", "invalid-alignment")

        match_output(
            cli,
            exit_code=1,
            exception_type=SettingsError,
            exception_pattern=exception_pattern,
            prog_name="test",
        )

    def test_attach_settings__persist(self, fake_settings_path: Path, runner: CliRunner):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel, persist=True)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            settings = get_settings(ctx, DefaultSettingsModel)
            settings.name = "hutt"
            settings.planet = "nal hutta"
            settings.is_humanoid = False
            settings.alignment = "evil"

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 0

        assert fake_settings_path.exists()
        data = json.loads(fake_settings_path.read_text())
        assert data == dict(
            name="hutt",
            planet="nal hutta",
            is_humanoid=False,
            alignment="evil",
        )

    def test_attach_settings__loads_settings_from_disk_if_file_exists(
        self,
        fake_settings_path: Path,
        runner: CliRunner,
    ):
        fake_settings_path.write_text(
            json.dumps(
                dict(
                    name="jawa",
                    planet="tatooine",
                    is_humanoid=True,
                    alignment="neutral",
                )
            )
        )
        cli = typer.Typer()

        @cli.command()
        @attach_settings(RequiredFieldsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            settings = get_settings(ctx, RequiredFieldsModel)
            assert settings.name == "jawa"
            assert settings.planet == "tatooine"
            assert settings.is_humanoid
            assert settings.alignment == "neutral"

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 0

    def test_attach_settings__show_no_footer(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel, show=True)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            pass

        expected_pattern = [
            "name.*jawa",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

    def test_attach_settings__adds_footer_if_persisted(self, fake_settings_path: Path):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel, persist=True, show=True)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            pass

        expected_pattern = [
            "name.*jawa",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
            f"saved to {str(fake_settings_path)[:40]}",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )


class TestWithParameters:
    def test_attach_settings__with_settings_parameter(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel, show=True)
        def noop(ctx: typer.Context, stuff: DefaultSettingsModel):  # pyright: ignore[reportUnusedFunction]
            assert stuff.name == "jawa"
            assert stuff.planet == "tatooine"
            assert stuff.is_humanoid
            assert stuff.alignment == "neutral"
            ctx_settings = get_settings(ctx, DefaultSettingsModel)
            assert ctx_settings is stuff

        expected_pattern = [
            "name.*jawa",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        match_help(cli, unwanted_pattern="stuff")


    def test_attach_settings__with_manager_parameter(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel, show=True)
        def noop(ctx: typer.Context, mgr: SettingsManager):  # pyright: ignore[reportUnusedFunction]
            settings: DefaultSettingsModel = cast(DefaultSettingsModel, mgr.settings_instance)
            assert settings.name == "jawa"
            assert settings.planet == "tatooine"
            assert settings.is_humanoid
            assert settings.alignment == "neutral"
            ctx_settings = get_settings(ctx, DefaultSettingsModel)
            assert ctx_settings is settings

        expected_pattern = [
            "name.*jawa",
            "planet.*tatooine",
            "is-humanoid.*True",
            "alignment.*neutral",
        ]
        match_output(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        match_help(cli, unwanted_pattern="mgr")


class TestGetSettings:
    def test_get_settings__extracts_settings_from_context(self, runner: CliRunner):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            settings = get_settings(ctx, DefaultSettingsModel)
            assert isinstance(settings, DefaultSettingsModel)
            assert settings.name == "jawa"
            assert settings.planet == "tatooine"
            assert settings.is_humanoid
            assert settings.alignment == "neutral"

        result = runner.invoke(cli, [], prog_name="test")
        assert result.exit_code == 0

    def test_get_settings__raises_error_with_no_attached_settings(self):
        cli = typer.Typer()

        @cli.command()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_settings(ctx, DefaultSettingsModel)

        match_output(
            cli,
            exception_type=SettingsError,
            exception_pattern="Settings are not bound to the context",
            exit_code=1,
            prog_name="test",
        )

    def test_get_settings__raises_error_with_wrong_settings_type(self):
        class NoMatchSettings(BaseModel):
            name: str = "jawa"
            planet: str = "tatooine"
            is_humanoid: bool = True
            alignment: str = "neutral"

        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_settings(ctx, NoMatchSettings)

        match_output(
            cli,
            exception_type=SettingsError,
            exception_pattern="instance doesn't match expected",
            exit_code=1,
            prog_name="test",
        )


class TestGetManager:
    def test_get_manager__extracts_settings_manager_from_context(self, fake_settings_path: Path):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert isinstance(manager, SettingsManager)
            assert manager.app_name == "test"
            assert manager.settings_model == DefaultSettingsModel
            assert manager.settings_path == fake_settings_path
            assert manager.invalid_warnings == {}
            assert isinstance(manager.settings_instance, DefaultSettingsModel)
            print("Passed!")

        match_output(
            cli,
            expected_pattern=["Passed"],
            exit_code=0,
            prog_name="test",
        )

    def test_get_manager__raises_exception_if_context_has_no_manager(self):
        cli = typer.Typer()

        @cli.command()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=SettingsError,
            exception_pattern="Settings are not bound to the context",
            exit_code=1,
            prog_name="test",
        )

    def test_get_manager__raises_exception_if_non_manager_retrieved(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(DefaultSettingsModel)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            ctx.obj.settings_manager = "Not a manager!"
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=SettingsError,
            exception_pattern="Item in user context at `settings_manager` was not a SettingsManager",
            exit_code=1,
            prog_name="test",
        )



from typing import Annotated

from pydantic import AnyHttpUrl, BaseModel, BeforeValidator

from typerdrive.client.base import ClientLogFunc, TyperdriveClient
from typerdrive.client.exceptions import ClientError


def pyright_safe_validator(value: str) -> str:
    AnyHttpUrl(value)
    return value


class ClientSpec(BaseModel):
    base_url: Annotated[str, BeforeValidator(pyright_safe_validator)]
    log_func: ClientLogFunc | None = None


class ClientManager:

    clients: dict[str, TyperdriveClient]

    def __init__(self):
        self.clients = {}

    def add_client(self, name: str, spec: ClientSpec) -> None:
        ClientError.require_condition(
            name not in self.clients,
            f"Client with name {name} already exists in context",
        )
        self.clients[name] = TyperdriveClient(log_func=spec.log_func, base_url=str(spec.base_url))

    def get_client(self, name: str) -> TyperdriveClient:
        return ClientError.enforce_defined(self.clients.get(name), f"No client named {name} found in context")



"""
This set of demos shows the use of the `attach_settings` decorator.
"""

import typer
from pydantic import BaseModel
from typerdrive.constants import Validation
from typerdrive.settings.attach import attach_settings, get_settings


def demo_1__attach_settings__basic():
    """
    This function demonstrates the use of the `attach_settings` decorator.
    This decorator allows you to attach a Pydantic model to a typer command.
    The settings are accessed by providing an argument with matching model type.
    Note that the settings argument name can be anything you like.
    In order to use this decorator, your command must take a `typer.Context`
    object as the first argument.
    """

    class ExampleSettings(BaseModel):
        name: str = "jawa"
        planet: str = "tatooine"
        alignment: str = "neutral"

    cli = typer.Typer()

    @cli.command()
    @attach_settings(ExampleSettings)
    def report(ctx: typer.Context, cfg: ExampleSettings):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        print(f"Look at this {cfg.name} from {cfg.planet}. It's soooo {cfg.alignment}!")

    cli()


def demo_2__attach_settings__enforce_validation():
    """
    This function demonstrates how the `attach_settings` decorator will,
    by default, require the settings to be valid before they can be
    attached to the app's context.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str = "tatooine"
        alignment: str = "neutral"

    cli = typer.Typer()

    @cli.command()
    @attach_settings(ExampleSettings)
    def report(ctx: typer.Context, cfg: ExampleSettings):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        print("We will never got to this line")

    cli()


def demo_3__attach_settings__allow_invalid():
    """
    This function demonstrates how the `validation` argument passed to the
    `attach_settings` decorator works. By default this flag is set to
    `Validation.BEFORE`, meaning the settings must be valid before it is
    attached to the app's context. If you set it to `Validation.NONE`, the
    settings will be bound even if there are invalid settings. This is
    important because your settings may have non-default settings that need
    to be set _through_ a command.
    """

    class ExampleSettings(BaseModel):
        name: str
        planet: str = "tatooine"
        alignment: str = "neutral"

    cli = typer.Typer()

    @cli.command()
    @attach_settings(ExampleSettings, validation=Validation.NONE)
    def report(ctx: typer.Context, cfg: ExampleSettings):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        print(f"Here are the settings that are missing a required field: {cfg}")

    cli()


def demo_4__attach_settings__access_through_context():
    """
    This function demonstrates how the settings can also be accessed through
    the Typer context if you do not want to use a settings parameter. To access
    settings with the context, use the `get_settings()` helper function.
    """

    class ExampleSettings(BaseModel):
        name: str = "jawa"
        planet: str = "tatooine"
        alignment: str = "neutral"

    cli = typer.Typer()

    @cli.command()
    @attach_settings(ExampleSettings)
    def report(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
        cfg = get_settings(ctx, ExampleSettings)
        print(f"Look at this {cfg.name} from {cfg.planet}. It's soooo {cfg.alignment}!")

    cli()



from pathlib import Path
from collections.abc import Generator

import pytest

from typerdrive.env import tweak_env


@pytest.fixture
def fake_cache_path(tmp_path: Path) -> Generator[Path, None, None]:
    fake_path = tmp_path / ".cache/test"
    fake_path.mkdir(parents=True)

    with tweak_env(HOME=str(tmp_path)):
        yield fake_path



"""
This set of demos shows the use of the `@attach_client()` decorator.
"""

from random import randint

import typer
from pydantic import BaseModel
from rich import print

from typerdrive.client.attach import attach_client
from typerdrive.client.base import TyperdriveClient
from typerdrive.settings.attach import attach_settings


def demo_1__attach_client__use_base_url_from_settings():
    """
    This function demonstrates how to use the `@attach_client()` decorator
    to attach a client to the app context using a base_url from the attached
    settings. The client provided is a TyperdriveClient that includes the
    enhanced "x" methods. In this example, `get_x()` is used to deserialize
    the response into a pydantic model. The `base_url` to use for the client
    is pulled from the matching field in the settings.
    """

    class SettingsModel(BaseModel):
        base_url: str = "https://swapi.info/api/people"

    class ResponseModel(BaseModel, extra="ignore"):
        name: str
        height: int
        mass: int
        birth_year: str
        gender: str

    cli = typer.Typer()

    @cli.command()
    @attach_settings(SettingsModel)
    @attach_client(swapi="base_url")
    def report(ctx: typer.Context, swapi: TyperdriveClient):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        person_id = randint(1, 10)
        print(f"Person {person_id} ->", swapi.get_x(f"{person_id}", response_model=ResponseModel))

    cli()


def demo_2__attach_client__use_explicit_base_url():
    """
    This function demonstrates how a client can be initialized without
    settings if the full base_url is provided.
    """

    class ResponseModel(BaseModel, extra="ignore"):
        name: str
        height: int
        mass: int
        birth_year: str
        gender: str

    cli = typer.Typer()

    @cli.command()
    @attach_client(swapi="https://swapi.info/api/people")
    def report(ctx: typer.Context, swapi: TyperdriveClient):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        person_id = randint(1, 10)
        print(f"Person {person_id} ->", swapi.get_x(f"{person_id}", response_model=ResponseModel))

    cli()


def demo_3__attach_client__multiple_clients():
    """
    This function demonstrates how multiple clients may be attached.
    The clients are independent of each other.
    """

    class SettingsModel(BaseModel):
        people_url: str = "https://swapi.info/api/people"
        planets_url: str = "https://swapi.info/api/planets"


    class PeopleResponse(BaseModel, extra="ignore"):
        name: str
        height: int
        mass: int
        birth_year: str
        gender: str


    class PlanetResponse(BaseModel, extra="ignore"):
        name: str
        climate: str
        terrain: str
        gravity: str
        population: int

    cli = typer.Typer()

    @cli.command()
    @attach_settings(SettingsModel)
    @attach_client(
        people="people_url",
        planets="planets_url",
    )
    def report(ctx: typer.Context, people: TyperdriveClient, planets: TyperdriveClient, id: int = 1):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        person_id = randint(1, 10)
        planet_id = randint(1, 10)

        print(f"Person {person_id} ->", people.get_x(f"{person_id}", response_model=PeopleResponse))
        print(f"Planet {planet_id} ->", planets.get_x(f"{planet_id}", response_model=PlanetResponse))

    cli()



import json
from pathlib import Path
from typing import Any, cast

from inflection import dasherize
from pydantic import BaseModel, ValidationError

from typerdrive.settings.exceptions import (
    SettingsInitError,
    SettingsResetError,
    SettingsSaveError,
    SettingsUnsetError,
    SettingsUpdateError,
)


def get_settings_path(app_name: str) -> Path:
    return Path.home() / ".local/share" / app_name / "settings.json"


class SettingsManager:
    def __init__(self, app_name: str, settings_model: type[BaseModel]):
        self.app_name: str = app_name
        self.settings_model: type[BaseModel] = settings_model
        self.settings_path: Path = get_settings_path(self.app_name)
        self.invalid_warnings: dict[str, str] = {}
        self.settings_instance: BaseModel
        with SettingsInitError.handle_errors("Failed to initialize settings"):
            settings_values: dict[str, Any] = {}
            if self.settings_path.exists():
                settings_values = json.loads(self.settings_path.read_text())
            try:
                self.settings_instance = self.settings_model(**settings_values)
            except ValidationError as err:
                self.settings_instance = self.settings_model.model_construct(**settings_values)
                self.set_warnings(err)

    def set_warnings(self, err: ValidationError):
        self.invalid_warnings = {}
        for data in err.errors():
            key: str = cast(str, data["loc"][0])
            message = data["msg"]
            self.invalid_warnings[key] = message

    def update(self, **settings_values: Any):
        with SettingsUpdateError.handle_errors("Failed to update settings"):
            combined_settings = {**self.settings_instance.model_dump(), **settings_values}
            try:
                self.settings_instance = self.settings_model(**combined_settings)
                self.invalid_warnings = {}
            except ValidationError as err:
                self.settings_instance = self.settings_model.model_construct(**combined_settings)
                self.set_warnings(err)

    def unset(self, *unset_keys: str):
        with SettingsUnsetError.handle_errors("Failed to remove keys"):
            settings_values = {k: v for (k, v) in self.settings_instance.model_dump().items() if k not in unset_keys}
            try:
                self.settings_instance = self.settings_model(**settings_values)
                self.invalid_warnings = {}
            except ValidationError as err:
                self.settings_instance = self.settings_model.model_construct(**settings_values)
                self.set_warnings(err)

    def reset(self):
        with SettingsResetError.handle_errors("Failed to reset settings"):
            try:
                self.settings_instance = self.settings_model()
                self.invalid_warnings = {}
            except ValidationError as err:
                self.settings_instance = self.settings_model.model_construct()
                self.set_warnings(err)

    def validate(self):
        self.settings_model(**self.settings_instance.model_dump())

    def pretty(self, with_style: bool = True) -> str:
        (bold_, _bold) = ("[bold]", "[/bold]") if with_style else ("", "")
        (red_, _red) = ("[red]", "[/red]") if with_style else ("", "")
        lines: list[str] = []
        parts: list[tuple[str, Any]] = []
        for field_name, field_value in self.settings_instance:
            if field_name == "invalid_warning":
                continue
            field_string = str(field_value)
            if field_name in self.invalid_warnings:
                field_string = f"{red_}{field_string}{_red}"
            parts.append((dasherize(field_name), field_string))

        max_field_len = max(len(field_name) for field_name, _ in parts)
        lines.extend(f"{bold_}{k:>{max_field_len}}{_bold} -> {v}" for k, v in parts)

        if self.invalid_warnings:
            lines.append("")
            lines.append(f"{red_}Settings are invalid:{_red}")
            lines.extend(f"{bold_}{k:>{max_field_len}}{_bold} -> {v}" for k, v in self.invalid_warnings.items())

        return "\n".join(lines)

    def save(self):
        with SettingsSaveError.handle_errors(f"Failed to save settings to {self.settings_path}"):
            self.settings_path.parent.mkdir(parents=True, exist_ok=True)
            self.settings_path.write_text(self.settings_instance.model_dump_json(indent=2))



from io import StringIO

from rich.console import Console
from rich.tree import Tree
from snick import strip_ansi_escape_sequences


def tree_to_text(tree: Tree) -> str:
    buffer = StringIO()
    console = Console(file=buffer, force_terminal=True, width=80)
    console.print(tree)
    return strip_ansi_escape_sequences(buffer.getvalue()).strip()



import typer
from pydantic import BaseModel

from typerdrive.client.base import TyperdriveClient
from typerdrive.client.attach import get_manager, attach_client
from typerdrive.client.exceptions import ClientError
from typerdrive.client.manager import ClientManager
from typerdrive.settings.attach import attach_settings

from tests.helpers import check_output, match_output


class ClientSettings(BaseModel):
    url_base: str = "https://the.force.io"


class TestAttachClient:
    def test_attach_client__adds_client_manager_to_context(self):
        cli = typer.Typer()

        @cli.command()
        @attach_client()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert isinstance(manager, ClientManager)
            assert manager.clients == {}
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")

    def test_attach_client__adds_a_client_using_explicit_base_url(self):
        cli = typer.Typer()

        @cli.command()
        @attach_client(jedi="https://the.force.io")
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert len(manager.clients) == 1
            client = manager.clients.get("jedi")
            assert client is not None
            assert client.base_url == "https://the.force.io"
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")

    def test_attach_client__adds_a_client_with_base_url_from_settings(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(ClientSettings)
        @attach_client(jedi="url_base")
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert len(manager.clients) == 1
            client = manager.clients.get("jedi")
            assert client is not None
            assert client.base_url == "https://the.force.io"
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")

    def test_attach_client__raises_an_error_if_invalid_url(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(ClientSettings)
        @attach_client(jedi="not-a-url")
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_manager(ctx)
            print("We should never get here")

        match_output(
            cli,
            unwanted_pattern="We should never get here",
            exception_type=ClientError,
            exception_pattern=r"Couldn't use base_url='not-a-url' for client",
            exit_code=1,
            prog_name="test",
        )

    def test_attach_client__adds_multiple_clients(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(ClientSettings)
        @attach_client(sith="https://the.dark.side", jedi="url_base")
        def noop(ctx: typer.Context): # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert len(manager.clients) == 2

            sith_client = manager.clients.get("sith")
            assert sith_client is not None
            assert sith_client.base_url == "https://the.dark.side"

            jedi_client = manager.clients.get("jedi")
            assert jedi_client is not None
            assert jedi_client.base_url == "https://the.force.io"
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")


class TestWithParameters:
    def test_attach_client__with_manager_parameter(self):
        cli = typer.Typer()

        @cli.command()
        @attach_client()
        def noop(ctx: typer.Context, mgr: ClientManager):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            assert isinstance(mgr, ClientManager)
            assert mgr.clients == {}
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")

    def test_attach_client__with_client_parameters(self):
        cli = typer.Typer()

        @cli.command()
        @attach_settings(ClientSettings)
        @attach_client(sith="https://the.dark.side", jedi="url_base")
        def noop(ctx: typer.Context, jedi: TyperdriveClient, sith: TyperdriveClient):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            assert sith is not None
            assert sith.base_url == "https://the.dark.side"

            assert jedi is not None
            assert jedi.base_url == "https://the.force.io"
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")


class TestGetManager:
    def test_get_manager__extracts_client_manager_from_context(self):
        cli = typer.Typer()

        @cli.command()
        @attach_client()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert isinstance(manager, ClientManager)
            assert manager.clients == {}
            print("Passed!")

        check_output(cli, prog_name="test", expected_substring="Passed!")

    def test_get_manager__raises_exception_if_context_has_no_manager(self):
        cli = typer.Typer()

        @cli.command()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=ClientError,
            exception_pattern=r"Client\(s\) are not bound to the context",
            escape_parens=True,
            exit_code=1,
            prog_name="test",
        )

    def test_get_manager__raises_exception_if_non_manager_retrieved(self):
        cli = typer.Typer()

        @cli.command()
        @attach_client()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            ctx.obj.client_manager = "Not a manager!"
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=ClientError,
            exception_pattern="Item in user context at `client_manager` was not a ClientManager",
            exit_code=1,
            prog_name="test",
        )



from enum import Flag, IntEnum, auto


class Validation(Flag):
    """
    Defines whether validation should happen "before", "after", "both", or "none"
    """

    BEFORE = auto()
    AFTER = auto()
    BOTH = BEFORE | AFTER
    NONE = 0


class ExitCode(IntEnum):
    """
    Maps exit codes for the application.
    """

    SUCCESS = 0
    GENERAL_ERROR = 1
    CANNOT_EXECUTE = 126
    INTERNAL_ERROR = 128



"""
This set of demos shows the use of the `handle_errors` decorator.
"""

import typer
from buzz import DoExceptParams

from typerdrive.exceptions import TyperdriveError, handle_errors


def demo_1__handle_errors__no_handler():
    """
    This function demonstrates what happens when no handler is installed.
    The output to the user is not nicely formatted. Additionally, in
    actual usage, it includes a stack trace. Generally end-users don't
    find stack-traces to be that helpful.
    """

    cli = typer.Typer()

    @cli.command()
    def report():  # pyright: ignore[reportUnusedFunction]
        raise TyperdriveError("Boring conversation anyway")

    cli()


def demo_2__handle_errors__basic():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to handle `TyperdriveError` exceptions (and all descendants). The
    decorator allows you to capture the errors and print a friendly message
    for the user instead of showing a stack trace.
    """

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?")
    def report():  # pyright: ignore[reportUnusedFunction]
        raise TyperdriveError("Boring conversation anyway")

    cli()


def demo_3__handle_errors__handle_specific_errors():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to handle specific exception types using the `handle_exc_class`
    parameter. Any instance of the supplied exception type will be caught,
    as will any descendant exceptions. You can also provide a group of
    exception types to handle by passing this parameter a tuple of exception
    types.
    """

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?", handle_exc_class=RuntimeError)
    def report():  # pyright: ignore[reportUnusedFunction]
        raise RuntimeError("Boring conversation anyway")

    cli()


def demo_4__handle_errors__ignore_specific_errors():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to ignore specific exception types while handling others using the
    `handle_exc_class` and `ignore_exc_class` parameters together. This
    is useful if you want to handle all descendants of a base class but
    ignore one specific descendant class. You can also provide a group of
    exception types to ignore by passing this parameter a tuple of exception
    types.
    """

    cli = typer.Typer()

    class BlasterRuntimeError(RuntimeError):
        pass

    @cli.command()
    @handle_errors(
        "What happened?",
        handle_exc_class=RuntimeError,
        ignore_exc_class=BlasterRuntimeError,
    )
    def report():  # pyright: ignore[reportUnusedFunction]
        raise BlasterRuntimeError("Boring conversation anyway")

    cli()


def demo_5__handle_errors__perform_task_on_error():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to add a task to perform when an exception is handled using the
    `do_except` parameter. The supplied function must take a single
    [`DoExceptParams`](https://dusktreader.github.io/py-buzz/reference/#buzz.tools.DoExceptParams)
    argument that carries essential information about the error.
    This kind of extra task is very useful if you want to log the error
    somewhere so that you can inspect it later.
    """

    def fake_logger(params: DoExceptParams):
        print(f"LOGGED: {params.final_message}")

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?", do_except=fake_logger)
    def report():  # pyright: ignore[reportUnusedFunction]
        raise TyperdriveError("Boring conversation anyway")

    cli()


def demo_6__handle_errors__perform_task_with_no_error():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to add a task to perform when no error is handled using the
    `do_else` parameter. The supplied function can't take any
    arguments. This kind of task is less useful than `do_except`,
    but you can find occasional uses for it. In this case, we will "log"
    a message that no error occurred.
    """

    def fake_logger():
        print("LOGGED: No errors occurred!")

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?", do_else=fake_logger)
    def report():  # pyright: ignore[reportUnusedFunction]
        print("We're all fine here now, thank you. How are you?")

    cli()


def demo_7__handle_errors__always_perform_task():
    """
    This function demonstrates how to use the `handle_errors` decorator
    to add a task to _always_ perform whether or not an error is handled
    with the `do_finally` parameter. The supplied function can't take any
    arguments. This kind of task will always be performed at the end of
    the `handle_errors` decorator whether _any_ kind of exception is
    raised or not.
    """

    def fake_logger():
        print("LOGGED: all done!")

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?", do_finally=fake_logger)
    def report():  # pyright: ignore[reportUnusedFunction]
        raise RuntimeError("Boring conversation anyway")

    cli()


def demo_8__handle_errors__set_a_subject_and_footer():
    """
    This function demonstrates how to use the `handle_errors` decorator
    with a custom subject and footer that will be displayed with the
    error message.
    """

    cli = typer.Typer()

    @cli.command()
    @handle_errors("What happened?")
    def report():  # pyright: ignore[reportUnusedFunction]
        raise TyperdriveError(
            "Boring conversation anyway",
            subject="slight weapons malfunction",
            footer="WE'RE GONNA HAVE COMPANY!",
        )

    cli()



from typerdrive.constants import ExitCode
from typerdrive.exceptions import TyperdriveError


class SettingsError(TyperdriveError):
    exit_code: ExitCode = ExitCode.GENERAL_ERROR


class SettingsInitError(SettingsError):
    pass


class SettingsUnsetError(SettingsError):
    pass


class SettingsResetError(SettingsError):
    pass


class SettingsUpdateError(SettingsError):
    pass


class SettingsSaveError(SettingsError):
    pass



from dataclasses import dataclass
from typing import cast, get_origin, get_args
from types import NoneType, UnionType

from buzz import require_condition
import typer

from typerdrive.cache.manager import CacheManager
from typerdrive.client.manager import ClientManager
from typerdrive.exceptions import ContextError
from typerdrive.settings.manager import SettingsManager


type TyperdriveManager = SettingsManager | CacheManager | ClientManager


@dataclass
class TyperdriveContext:
    settings_manager: SettingsManager | None = None
    cache_manager: CacheManager | None = None
    client_manager: ClientManager | None = None


def get_user_context(ctx: typer.Context):
    if not ctx.obj:
        ctx.obj = TyperdriveContext()
    return ctx.obj


def to_context(ctx: typer.Context, name: str, val: TyperdriveManager) -> None:
    user_context = get_user_context(ctx)
    field_type = TyperdriveContext.__dataclass_fields__[name].type

    if get_origin(field_type) is UnionType:
        defined_types = [t for t in get_args(field_type) if t is not NoneType]
        require_condition(
            len(defined_types) == 1,
            "PANIC! TyperdriveContext fields must only have one type or None.",
            raise_exc_class=RuntimeError,
        )
        field_type = defined_types[0]

    # TODO: Get the type hinting on the next line right.
    ContextError.ensure_type(val, field_type, "Value is not of type any of the union types")  # type: ignore[arg-type]

    setattr(user_context, name, val)


def from_context(ctx: typer.Context, name: str) -> TyperdriveManager:
    user_context = get_user_context(ctx)
    return ContextError.enforce_defined(getattr(user_context, name), f"{name} is not bound to context")


def get_app_name(ctx: typer.Context) -> str:
    if ctx.parent:
        return get_app_name(cast(typer.Context, ctx.parent))
    return ContextError.enforce_defined(ctx.info_name, "typerdrive requires the app to have a name")



from typerdrive.settings.attach import attach_settings, get_settings
from typerdrive.settings.manager import SettingsManager, get_settings_path
from typerdrive.settings.commands import add_bind, add_update, add_unset, add_reset, add_show, add_settings_subcommand


__all__ = [
    "SettingsManager",
    "add_bind",
    "add_reset",
    "add_settings_subcommand",
    "add_show",
    "add_unset",
    "add_update",
    "attach_settings",
    "get_settings",
    "get_settings_path",
]



from enum import StrEnum, auto
from typing import Annotated
from collections.abc import Callable

import snick
import typer
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Confirm

from typerdrive_demo.helpers import run_demo, get_demo_functions


class Feature(StrEnum):
    attach_settings = auto()
    settings_commands = auto()
    attach_cache = auto()
    cache_commands = auto()
    handle_errors = auto()
    attach_client = auto()
    typerdrive_client = auto()


def start(
    feature: Annotated[
        Feature | None,
        typer.Option(
            help="The feature to demo. If not provided, demo ALL",
        ),
    ] = None,
):
    """
    This cli app will demo the features of `typerdrive`!

    Note: If no features are selected for the demo, all of them will be shown.
    """
    features: list[Feature]
    if feature is None:
        features = [f for f in Feature]
    else:
        features = [feature]

    feature_map: dict[Feature, list[Callable[..., None]]] = {}
    for feature in features:
        feature_map[feature] = get_demo_functions(feature)

    override_label_map: dict[Feature, str] = {}

    greeting_lines = [
        "You are viewing the `typerdrive` demo!",
        "",
        "This program will show you the different features available in `typerdrive` and what it's like to use them",
        "",
        "The following features will be included:",
    ]
    for feature in features:
        label = override_label_map.get(feature, f"{feature}()")
        greeting_lines.append(f"- `{label}`")
        for demo in feature_map[feature]:
            greeting_lines.append(f"  - `{demo.__name__}()`")

    console = Console()
    console.clear()
    console.print(
        Panel(
            Markdown(snick.conjoin(*greeting_lines)),
            padding=1,
            title="[green]Welcome to typerdrive![/green]",
            title_align="left",
            subtitle="[blue]https://github.com/dusktreader/typerdrive[/blue]",
            subtitle_align="left",
        )
    )
    console.print()
    console.print()
    further: bool = Confirm.ask("Would you like to continue?", default=True)
    if not further:
        return

    for feature, demos in feature_map.items():
        for demo in demos:
            further = run_demo(demo, console, override_label=override_label_map.get(feature))
            if not further:
                break

    console.clear()
    console.print(
        Panel(
            Markdown(
                snick.dedent(
                    """
                    Thanks for checking out `typerdrive`!

                    I hope these features will be useful for you.

                    If you would like to learn more, please check out the
                    [documentation site](https://dusktreader.github.io/typerdrive/)
                    """
                ),
            ),
            padding=1,
            title="[green]Thanks![/green]",
            title_align="left",
            subtitle="[blue]https://github.com/dusktreader/typerdrive[/blue]",
            subtitle_align="left",
        )
    )
    console.print()
    console.print()


def main():
    typer.run(start)



from collections.abc import Callable
from functools import wraps
from typing import Any, ParamSpec, TypeVar

from buzz import Buzz, DoExceptParams, reformat_exception, get_traceback
from rich import traceback
import snick
import typer

from typerdrive.constants import ExitCode
from typerdrive.format import terminal_message


traceback.install()


class TyperdriveError(Buzz):
    subject: str | None = None
    footer: str | None = None
    exit_code: ExitCode = ExitCode.GENERAL_ERROR
    details: Any | None = None

    def __init__(
        self,
        *args: Any,
        subject: str | None = None,
        footer: str | None = None,
        details: Any | None = None,
        exit_code: ExitCode | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if subject:
            self.subject = subject
        if footer:
            self.footer = footer
        if details:
            self.details = details
        if exit_code:
            self.exit_code = exit_code


class ContextError(TyperdriveError):
    exit_code: ExitCode = ExitCode.INTERNAL_ERROR


class BuildCommandError(TyperdriveError):
    exit_code: ExitCode = ExitCode.INTERNAL_ERROR


P = ParamSpec("P")
T = TypeVar("T")
WrappedFunction = Callable[P, T]


def handle_errors(
    base_message: str,
    *,
    handle_exc_class: type[Exception] | tuple[type[Exception], ...] = TyperdriveError,
    ignore_exc_class: type[Exception] | tuple[type[Exception], ...] | None = None,
    do_except: Callable[[DoExceptParams], None] | None = None,
    do_else: Callable[[], None] | None = None,
    do_finally: Callable[[], None] | None = None,
    unwrap_message: bool = True,
    debug: bool = False,
) -> Callable[[WrappedFunction[P, T]], WrappedFunction[P, T]]:
    class _DefaultIgnoreException(Exception):
        """
        Define a special exception class to use for the default ignore behavior.

        Basically, this exception type can't be extracted from this method (easily), and thus could never actually
        be raised in any other context. This is only created here to preserve the `try/except/except/else/finally`
        structure.
        """

    ignore_exc_class = _DefaultIgnoreException if ignore_exc_class is None else ignore_exc_class

    def _decorate(func: WrappedFunction[P, T]) -> WrappedFunction[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return_value: T | None = None
            try:
                return_value = func(*args, **kwargs)
            except ignore_exc_class:
                raise
            except handle_exc_class as err:
                try:
                    final_message = reformat_exception(base_message, err)
                except Exception as msg_err:
                    raise RuntimeError(f"Failed while formatting message: {repr(msg_err)}")

                trace = get_traceback()

                if do_except:
                    do_except(
                        DoExceptParams(
                            err=err,
                            base_message=base_message,
                            final_message=final_message,
                            trace=trace,
                        )
                    )

                subject: str | None = base_message
                footer: str | None = None
                message: str

                exit_code: int = ExitCode.GENERAL_ERROR
                if isinstance(err, TyperdriveError):
                    if err.subject:
                        subject = err.subject
                    if err.footer:
                        footer = err.footer
                    if debug:
                        message = err.message
                    else:
                        message = err.base_message or err.message
                    if unwrap_message:
                        message = snick.unwrap(message)
                    exit_code = err.exit_code
                else:
                    message = str(err)

                terminal_message(
                    message,
                    subject=f"[red]{subject}[/red]",
                    footer=footer,
                    error=True,
                )

                raise typer.Exit(code=exit_code)

            else:
                if do_else:
                    do_else()
                return return_value

            finally:
                if do_finally:
                    do_finally()

        return wrapper

    return _decorate



from collections.abc import Callable
from functools import wraps
from typing import Concatenate, ParamSpec, TypeVar, Annotated, Any

import humanize
import typer

from typerdrive.cache.exceptions import CacheError
from typerdrive.cache.manager import CacheManager
from typerdrive.cloaked import CloakingDevice
from typerdrive.context import from_context, to_context, get_app_name
from typerdrive.format import terminal_message


def get_manager(ctx: typer.Context) -> CacheManager:
    with CacheError.handle_errors(
        "Cache is not bound to the context. Use the @attach_cache() decorator"
    ):
        mgr: Any = from_context(ctx, "cache_manager")
    return CacheError.ensure_type(
        mgr,
        CacheManager,
        "Item in user context at `cache_manager` was not a CacheManager",
    )


P = ParamSpec("P")
T = TypeVar("T")
ContextFunction = Callable[Concatenate[typer.Context, P], T]


def attach_cache(show: bool = False) -> Callable[[ContextFunction[P, T]], ContextFunction[P, T]]:
    def _decorate(func: ContextFunction[P, T]) -> ContextFunction[P, T]:

        manager_param_key: str | None = None
        for key in func.__annotations__.keys():
            if func.__annotations__[key] is CacheManager:
                func.__annotations__[key] = Annotated[CacheManager | None, CloakingDevice]
                manager_param_key = key

        @wraps(func)
        def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> T:
            manager: CacheManager = CacheManager(get_app_name(ctx))
            to_context(ctx, "cache_manager", manager)

            if manager_param_key:
                kwargs[manager_param_key] = manager

            ret_val = func(ctx, *args, **kwargs)

            if show:
                cache_info = manager.pretty()
                human_size = humanize.naturalsize(cache_info.total_size)
                terminal_message(
                    cache_info.tree,
                    subject="Current cache",
                    footer=f"Storing {human_size} in {cache_info.file_count} files",
                )
            return ret_val

        return wrapper

    return _decorate



from pathlib import Path

import typer

from typerdrive.cache.commands import add_clear, add_show, add_cache_subcommand
from typerdrive.cache.manager import CacheManager
from typerdrive.constants import ExitCode

from tests.helpers import match_help, match_output


class TestClear:
    def test_clear__removes_a_single_entry_with_path_param(self, fake_cache_path: Path):
        cli = typer.Typer()
        add_clear(cli)
        full_path = fake_cache_path / "jawa/ewok.txt"
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_text("hive of scum and villainy")

        match_output(
            cli,
            "--path=jawa/ewok.txt",
            expected_pattern="Cleared entry at cache target jawa/ewok.txt",
            prog_name="test",
        )

        assert not full_path.exists()

    def test_clear__raises_exception_if_no_item_is_found_at_path(self):
        cli = typer.Typer()

        add_clear(cli)

        match_output(
            cli,
            "--path=jawa/ewok.txt",
            exit_code=ExitCode.GENERAL_ERROR,
            expected_pattern="Failed to clear cache target jawa/ewok.txt",
            prog_name="test",
        )

    def test_clear__help_text(self):
        cli = typer.Typer()

        add_clear(cli)

        expected_pattern = [
            "Options",
            r"--path TEXT Clear only the entry .* \[default: None\]",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

    def test_clear_all__removes_all_cached_files_with_no_path_arg(self, fake_cache_path: Path):
        cli = typer.Typer()
        add_clear(cli)

        manager: CacheManager = CacheManager("test")
        manager.store_text("hive of scum and villainy", "jawa/ewok")
        manager.store_bytes(b"that's no moon", "hutt")

        match_output(
            cli,
            expected_pattern="Cleared all 2 files from cache",
            input="y",
            prog_name="test",
        )

        assert not (fake_cache_path / "jawa/ewok").exists()
        assert not (fake_cache_path / "hutt").exists()

        assert len([p for p in fake_cache_path.iterdir()]) == 0

    def test_clear_all__does_nothing_if_confirmation_not_provided(self, fake_cache_path: Path):
        cli = typer.Typer()
        add_clear(cli)

        manager: CacheManager = CacheManager("test")
        manager.store_text("hive of scum and villainy", "jawa/ewok")
        manager.store_bytes(b"that's no moon", "hutt")

        match_output(
            cli,
            exit_code=1,
            input="n",
            prog_name="test",
        )

        assert (fake_cache_path / "jawa/ewok").exists()
        assert (fake_cache_path / "hutt").exists()

        assert len([p for p in fake_cache_path.iterdir()]) == 2


class TestShow:
    def test_show__shows_cache_directory_tree(self, fake_cache_path: Path):
        manager: CacheManager = CacheManager("test")
        manager.store_text("hive of scum and villainy", "jawa/ewok.txt")
        manager.store_bytes(b"that's no moon", "hutt")

        cli = typer.Typer()

        add_show(cli)

        expected_pattern = f"""
            📂 {fake_cache_path}
            ├── 📂 jawa
            │   └── 📄 ewok.txt (25 Bytes)
            └── 📄 hutt (14 Bytes)
        """
        match_output(
            cli,
            expected_pattern=expected_pattern,
            escape_parens=True,
            prog_name="test",
        )


class TestSubcommand:
    def test_add_cache_subcommand(self):
        cli = typer.Typer()

        add_cache_subcommand(cli)

        expected_pattern = [
            r"cache.*Manage cache for the app",
        ]
        match_help(
            cli,
            expected_pattern=expected_pattern,
            prog_name="test",
        )

        expected_pattern = [
            "clear",
            "show",
        ]
        match_output(
            cli,
            "cache",
            "--help",
            exit_code=0,
            expected_pattern=expected_pattern,
            prog_name="test",
        )



from typing import Annotated

import typer
from pydantic import BaseModel

from typerdrive.client.base import TyperdriveClient
from typerdrive.client.attach import attach_client
from typerdrive.format import terminal_message
from typerdrive.settings.attach import attach_settings
from typerdrive.exceptions import handle_errors


class SettingsModel(BaseModel):
    people_url: str = "https://swapi.info/api/people"
    planets_url: str = "https://swapi.info/api/planets"


class PeopleResponse(BaseModel, extra="ignore"):
    name: str
    height: int
    birth_year: str
    gender: str


class PlanetResponse(BaseModel, extra="ignore"):
    name: str
    climate: str
    terrain: str
    gravity: str
    population: int


cli = typer.Typer()


@cli.command()
@handle_errors("Lookup on SWAPI failed!")
@attach_settings(SettingsModel)
@attach_client(people="people_url", planets="planets_url")
def report(
    ctx: typer.Context,
    people: TyperdriveClient,
    planets: TyperdriveClient,
    person_id: Annotated[int, typer.Option(help="The ID of the person to look up")] = 1,
    planet_id: Annotated[int, typer.Option(help="The ID of the planet to look up")] = 1,
):  # pyright: ignore[reportUnusedParameter]
    terminal_message(
        str(people.get_x(f"{person_id}", expected_status=200, response_model=PeopleResponse)),
        subject=f"Person {person_id}",
        footer=f"Fetched from {people.base_url}{person_id}"
    )
    terminal_message(
        str(planets.get_x(f"{planet_id}", expected_status=200, response_model=PlanetResponse)),
        subject=f"Planet {planet_id}",
        footer=f"Fetched from {planets.base_url}{planet_id}"
    )


if __name__ == "__main__":
    cli()



import logging
from typing import Any, Callable

import pydantic
from httpx import Client, URL, RequestError

from typerdrive.client.exceptions import ClientError


ClientLogFunc = Callable[[str], None]


class TyperdriveClient(Client):
    log_func: ClientLogFunc

    def __init__(self, *args: Any, log_func: ClientLogFunc | None = None, **kwargs: Any):
        super().__init__(*args, **kwargs)

        if not log_func:
            logger: logging.Logger = logging.getLogger("typerdrive.client")
            self.log_func = logger.debug
        else:
            self.log_func = log_func

    def request_x[RM: pydantic.BaseModel](
        self,
        method: str,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:

        self.log_func(f"Processing {method} request to {url}")

        if param_obj is not None:
            self.log_func(f"Unpacking {param_obj=} to url params")

            ClientError.require_condition(
                "params" not in request_kwargs,
                "'params' not allowed when using param_obj",
            )
            with ClientError.handle_errors("Param data could not be deserialized for http request"):
                request_kwargs["params"] = param_obj.model_dump(mode="json")

        if body_obj is not None:
            self.log_func(f"Unpacking {body_obj=} to request body")

            ClientError.require_condition(
                all(k not in request_kwargs for k in ["data", "json", "content"]),
                "'data', 'json' and 'content' not allowed when using body_obj",
            )
            with ClientError.handle_errors("Request body data could not be deserialized for http request"):
                request_kwargs["content"] = body_obj.model_dump_json()
                request_kwargs["headers"] = {"Content-Type": "application/json"}

        with ClientError.handle_errors(
            "Communication with the API failed",
            handle_exc_class=RequestError,
        ):
            self.log_func("Issuing request")
            response = self.request(method, url, **request_kwargs)

        if expected_status is not None:
            self.log_func(f"Checking response for {expected_status=}")
            ClientError.require_condition(
                expected_status == response.status_code,
                "Got an unexpected status code: Expected {}, got {} -- {}".format(
                    expected_status, response.status_code, response.reason_phrase
                ),
                raise_kwargs=dict(details=response.text),
            )

        if not expect_response:
            self.log_func(f"Skipping response processing due to {expect_response=}")
            return response.status_code

        with ClientError.handle_errors("Failed to unpack JSON from response"):
            self.log_func("Parsing JSON from response")
            data: dict[str, Any] = response.json()

        if not response_model:
            self.log_func("Returning raw data due to no response model being supplied")
            return data

        with ClientError.handle_errors("Unexpected data in response"):
            self.log_func(f"Serializing response as {response_model.__name__}")
            return response_model(**data)

    def get_x[RM: pydantic.BaseModel](
        self,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:
        return self.request_x(
            "GET",
            url,
            param_obj=param_obj,
            body_obj=body_obj,
            expected_status=expected_status,
            expect_response=expect_response,
            response_model=response_model,
            **request_kwargs
        )

    def post_x[RM: pydantic.BaseModel](
        self,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:
        return self.request_x(
            "POST",
            url,
            param_obj=param_obj,
            body_obj=body_obj,
            expected_status=expected_status,
            expect_response=expect_response,
            response_model=response_model,
            **request_kwargs
        )

    def put_x[RM: pydantic.BaseModel](
        self,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:
        return self.request_x(
            "PUT",
            url,
            param_obj=param_obj,
            body_obj=body_obj,
            expected_status=expected_status,
            expect_response=expect_response,
            response_model=response_model,
            **request_kwargs
        )

    def patch_x[RM: pydantic.BaseModel](
        self,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:
        return self.request_x(
            "PATCH",
            url,
            param_obj=param_obj,
            body_obj=body_obj,
            expected_status=expected_status,
            expect_response=expect_response,
            response_model=response_model,
            **request_kwargs
        )

    def delete_x[RM: pydantic.BaseModel](
        self,
        url: URL | str,
        *,
        param_obj: pydantic.BaseModel | None = None,
        body_obj: pydantic.BaseModel | None = None,
        expected_status: int | None = None,
        expect_response: bool = True,
        response_model: type[RM] | None = None,
        **request_kwargs: Any,
    ) -> RM | int | dict[str, Any]:
        return self.request_x(
            "DELETE",
            url,
            param_obj=param_obj,
            body_obj=body_obj,
            expected_status=expected_status,
            expect_response=expect_response,
            response_model=response_model,
            **request_kwargs
        )




from collections.abc import Callable
from functools import wraps
from typing import Concatenate, ParamSpec, TypeVar, Annotated, Any

import typer
from pydantic import BaseModel
from typer_repyt.constants import Sentinel

from typerdrive.constants import Validation
from typerdrive.context import from_context, to_context, get_app_name
from typerdrive.format import terminal_message
from typerdrive.cloaked import CloakingDevice
from typerdrive.settings.exceptions import SettingsError
from typerdrive.settings.manager import SettingsManager


def get_settings[ST: BaseModel](
    ctx: typer.Context,
    type_hint: type[ST],
) -> ST:
    return SettingsError.ensure_type(
        get_manager(ctx).settings_instance,
        type_hint,
        f"Settings instance doesn't match expected {type_hint=}",
    )


def get_settings_value(ctx: typer.Context, settings_key: str) -> Any:
    instance: BaseModel = get_manager(ctx).settings_instance
    ret = getattr(instance, settings_key, Sentinel.MISSING)
    SettingsError.require_condition(
        ret is not Sentinel.MISSING,
        f"Settings instance doesn't have field {settings_key}",
    )
    return ret


def get_manager(ctx: typer.Context) -> SettingsManager:
    with SettingsError.handle_errors(
        "Settings are not bound to the context. Use the @attach_settings() decorator"
    ):
        mgr: Any = from_context(ctx, "settings_manager")
    return SettingsError.ensure_type(
        mgr,
        SettingsManager,
        "Item in user context at `settings_manager` was not a SettingsManager",
    )


P = ParamSpec("P")
T = TypeVar("T")
ContextFunction = Callable[Concatenate[typer.Context, P], T]


def attach_settings(
    settings_model: type[BaseModel],
    *,
    validation: Validation = Validation.BEFORE,
    persist: bool = False,
    show: bool = False,
) -> Callable[[ContextFunction[P, T]], ContextFunction[P, T]]:
    def _decorate(func: ContextFunction[P, T]) -> ContextFunction[P, T]:

        manager_param_key: str | None = None
        settings_param_key: str | None = None
        for key in func.__annotations__.keys():
            if func.__annotations__[key] is settings_model:
                func.__annotations__[key] = Annotated[settings_model | None, CloakingDevice]
                settings_param_key = key
            elif func.__annotations__[key] is SettingsManager:
                func.__annotations__[key] = Annotated[SettingsManager | None, CloakingDevice]
                manager_param_key = key

        # TODO: Figure out how we can make the ctx param optional for the wrapped function
        @wraps(func)
        def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> T:
            manager: SettingsManager = SettingsManager(get_app_name(ctx), settings_model)

            if validation & Validation.BEFORE:
                SettingsError.require_condition(
                    len(manager.invalid_warnings) == 0,
                    f"Initial settings are invalid: {manager.invalid_warnings}",
                )
            to_context(ctx, "settings_manager", manager)

            if settings_param_key:
                kwargs[settings_param_key] = manager.settings_instance

            if manager_param_key:
                kwargs[manager_param_key] = manager

            ret_val = func(ctx, *args, **kwargs)

            if validation & Validation.AFTER:
                with SettingsError.handle_errors("Final settings are invalid"):
                    manager.validate()

            if persist:
                manager.save()
            if show:
                terminal_message(
                    manager.pretty(),
                    subject="Current settings",
                    footer=f"saved to {manager.settings_path}" if persist else None,
                )

            return ret_val

        return wrapper

    return _decorate



from pydantic import ValidationError
import pytest

from typerdrive.client.base import TyperdriveClient
from typerdrive.client.exceptions import ClientError
from typerdrive.client.manager import ClientManager, ClientSpec


class TestClientSpec:
    def test_accepts_valid_url(self):
        ClientSpec(base_url="https://the.force.io")

    def test_raises_error_on_invalid_url(self):
        with pytest.raises(ValidationError, match="Input should be a valid URL"):
            ClientSpec(base_url="not-a-url")


class TestClientManager:
    def test_inti__initializes_client_dict(self):
        manager = ClientManager()
        assert manager.clients == {}

    def test_add_client__adds_a_client(self):
        manager = ClientManager()
        spec = ClientSpec(base_url="https://the.force.io")
        manager.add_client("test", spec)
        assert len(manager.clients) == 1
        client = manager.clients["test"]
        assert str(client.base_url) == "https://the.force.io"

    def test_add_client__raises_error_if_client_already_exists(self):
        manager = ClientManager()
        manager.clients["test"] = TyperdriveClient()
        with pytest.raises(ClientError, match="Client with name test already exists in context"):
            manager.add_client("test", ClientSpec(base_url="https://the.force.io"))

    def test_get_client__fetches_a_client_by_name(self):
        manager = ClientManager()
        manager.clients["test"] = TyperdriveClient(base_url="https://the.force.io")
        client = manager.get_client("test")
        assert client.base_url == "https://the.force.io"

    def test_get_client__raises_an_error_if_the_client_does_not_exist(self):
        manager = ClientManager()
        with pytest.raises(ClientError, match="No client named test found in context"):
            manager.get_client("test")



import os
from contextlib import contextmanager


@contextmanager
def tweak_env(**kwargs: str):
    old_vals: dict[str, str] = {}
    del_vals: list[str] = []
    for key, val in kwargs.items():
        old_val = os.environ.get(key, None)
        if old_val:
            old_vals[key] = old_val
        else:
            del_vals.append(key)
        os.environ[key] = val

    yield

    for key, val in old_vals.items():
        os.environ[key] = val
    for key in del_vals:
        del os.environ[key]



from collections.abc import Callable
from functools import wraps
from typing import Annotated, Any, Concatenate, ParamSpec, TypeVar

import typer
from pydantic import BaseModel

from typerdrive.client.base import TyperdriveClient, ClientLogFunc
from typerdrive.client.exceptions import ClientError
from typerdrive.client.manager import ClientManager, ClientSpec
from typerdrive.cloaked import CloakingDevice
from typerdrive.context import from_context, to_context
from typerdrive.settings.attach import get_manager as get_settings_manager
from typerdrive.settings.exceptions import SettingsError


def get_manager(ctx: typer.Context) -> ClientManager:
    with ClientError.handle_errors(
        "Client(s) are not bound to the context. Use the @attach_client() decorator"
    ):
        mgr: Any = from_context(ctx, "client_manager")
    return ClientError.ensure_type(
        mgr,
        ClientManager,
        "Item in user context at `client_manager` was not a ClientManager",
    )


def get_client(ctx: typer.Context, name: str) -> TyperdriveClient:
    return get_manager(ctx).get_client(name)


P = ParamSpec("P")
T = TypeVar("T")
ContextFunction = Callable[Concatenate[typer.Context, P], T]


def attach_client(
    log_func: ClientLogFunc | None = None,
    **client_urls_or_settings_keys: str,
) -> Callable[[ContextFunction[P, T]], ContextFunction[P, T]]:
    def _decorate(func: ContextFunction[P, T]) -> ContextFunction[P, T]:

        manager_param_key: str | None = None
        client_param_keys: list[str] = []
        for key in func.__annotations__.keys():
            if func.__annotations__[key] is TyperdriveClient:
                if key in client_urls_or_settings_keys:
                    func.__annotations__[key] = Annotated[TyperdriveClient | None, CloakingDevice]
                    client_param_keys.append(key)

            elif func.__annotations__[key] is ClientManager:
                func.__annotations__[key] = Annotated[ClientManager | None, CloakingDevice]
                manager_param_key = key

        @wraps(func)
        def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> T:
            manager = ClientManager()

            settings: BaseModel | None
            try:
                settings = get_settings_manager(ctx).settings_instance
            except SettingsError:
                settings = None

            for (name, url_or_settings_key) in client_urls_or_settings_keys.items():
                base_url: str = getattr(settings, url_or_settings_key, url_or_settings_key)
                with ClientError.handle_errors(
                    f"Couldn't use {base_url=} for client. If using a settings key, make sure settings are attached."
                ):
                    client_spec = ClientSpec(base_url=base_url, log_func=log_func)
                manager.add_client(name, client_spec)

            to_context(ctx, "client_manager", manager)

            for key in client_param_keys:
                kwargs[key] = manager.get_client(key)

            if manager_param_key:
                kwargs[manager_param_key] = manager

            return func(ctx, *args, **kwargs)

        return wrapper

    return _decorate



from random import choice

import typer
from rich import print
from rich.panel import Panel
from snick import unwrap

from typerdrive.cache.commands import add_cache_subcommand
from typerdrive.cache.attach import attach_cache
from typerdrive.cache.exceptions import CacheError
from typerdrive.cache.manager import CacheManager
from typerdrive.env import tweak_env


cli = typer.Typer()
add_cache_subcommand(cli)

speeches = dict(
    yoda=unwrap(
        """
        Size matters not. Look at me. Judge me by my size, do you? Hmm? Hmm. And well you should not. For my ally is the
        Force, and a powerful ally it is. Life creates it, makes it grow. Its energy surrounds us and binds us. Luminous
        beings are we, not this crude matter. You must feel the Force around you; here, between you, me, the tree, the
        rock, everywhere, yes. Even between the land and the ship.
        """
    ),
    leia=unwrap(
        """
        General Kenobi. Years ago you served my father in the Clone Wars. Now he begs you to help him in his struggle
        against the Empire. I regret that I am unable to present my father's request to you in person, but my ship has
        fallen under attack, and now I'm afraid my mission to bring you to Alderaan has failed.  I have placed
        information vital of the survival of the Rebellion into the memory systems of this R2 unit.My father will know
        how to retrieve it. You must see this droid safely delivered to him on Alderaan. This is our most desperate
        hour. Help me, Obi-Wan Kenobi. You're my only hope.
        """
    ),
    han=unwrap(
        """
        Kid, I've flown from one side of this galaxy to the other; I've seen a lot of strange stuff. But I've never seen
        anything to make me believe that there's one all-powerful Force controlling everything. There's no mystical
        energy field that controls my destiny. Anyway, it's all a lot of simple tricks and nonsense.
        """
    )
)


@cli.command()
@attach_cache()
def report(ctx: typer.Context, manager: CacheManager):
    speaker = choice(list(speeches.keys()))
    path = f"{speaker}/speech.txt"
    used_cache = False
    text: str
    try:
        text = manager.load_text(path)
    except CacheError:
        print(f"[red]Cache miss![/red] Loading text for {speaker}")
        text = speeches[speaker]
    else:
        print(f"[green]Cache hit![/green] Loaded text from cache target [yellow]{path}[/yellow]")
        used_cache = True

    with tweak_env(COLUMNS="80"):
        print(Panel(text))

    if not used_cache:
        manager.store_text(text, path)
        print(f"Stored text at cache target [yellow]{path}[/yellow]")


if __name__ == "__main__":
    cli()



from typerdrive.constants import ExitCode
from typerdrive.exceptions import TyperdriveError


class CacheError(TyperdriveError):
    exit_code: ExitCode = ExitCode.GENERAL_ERROR


class CacheInitError(CacheError):
    pass


class CacheStoreError(CacheError):
    pass


class CacheClearError(CacheError):
    pass


class CacheLoadError(CacheError):
    pass



"""
This set of demos shows the use of the `attach_cache` decorator.
"""

import json

import typer
from rich import print

from typerdrive.cache.manager import get_cache_path, CacheManager
from typerdrive.cache.attach import attach_cache, get_manager


def demo_1__attach_cache__storing_data():
    """
    This function demonstrates how to use the `attach_cache` decorator
    to store data to the cache. The decorator gives you to access the
    `CacheManager` within a typer command by adding a parameter with
    `CacheManager` type. Note that the manager argument name can be
    anything you like. The `CacheManager` can store bytes, text, or
    JSON data in the cache. In order to use this decorator, your
    command must take a `typer.Context` object as the first argument.
    """

    cli = typer.Typer()

    @cli.command()
    @attach_cache(show=True)
    def report(ctx: typer.Context, manager: CacheManager):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        manager.store_text("Utinni!", "jawa.txt")
        manager.store_bytes(b"Yub Nub!", "ewok")
        manager.store_json(
            dict(
                full_name="Jabba Desilijic Tiure",
                age=604,
                quote="beeska chata wnow kong bantha poodoo!",
            ),
            "hutt/jabba.json",
        )

    cli()


def demo_2__attach_cache__loading_data():
    """
    This function demonstrates how to use the `attach_cache` decorator
    to load data from the cache. Because the demos use a temporary
    directory, the cache from the previous command is no longer
    available, so this function pre-seeds the data.
    """

    cache_path = get_cache_path("demo_2__attach_cache__loading_data")
    jawa_path = cache_path / "jawa.txt"
    jawa_path.parent.mkdir(parents=True, exist_ok=True)
    jawa_path.write_text("Utinni!")
    ewok_path = cache_path / "ewok"
    ewok_path.parent.mkdir(parents=True, exist_ok=True)
    ewok_path.write_bytes(b"Yub Nub!")
    jabba_path = cache_path / "hutt/jabba.json"
    jabba_path.parent.mkdir(parents=True, exist_ok=True)
    jabba_path.write_text(
        json.dumps(
            dict(
                full_name="Jabba Desilijic Tiure",
                age=604,
                quote="beeska chata wnow kong bantha poodoo!",
            ),
        ),
    )

    cli = typer.Typer()

    @cli.command()
    @attach_cache(show=True)
    def report(ctx: typer.Context, manager: CacheManager):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
        jawa_quote = manager.load_text("jawa.txt")
        ewok_quote = manager.load_bytes("ewok").decode("utf-8")
        jabba_quote = manager.load_json("hutt/jabba.json")["quote"]
        print(f"The jawa says: '{jawa_quote}'")
        print(f"The ewok says: '{ewok_quote}'")
        print(f"Jabba the Hutt says: '{jabba_quote}'")

    cli()


def demo_4__attach_settings__access_through_context():
    """
    This function demonstrates how the cache manager can also be accessed
    through the Typer context if you do not want to use a `CacheManager`
    parameter. To access the manager with the context, use the
    `get_manager()` helper function
    """

    cli = typer.Typer()

    @cli.command()
    @attach_cache(show=True)
    def report(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
        manager = get_manager(ctx)
        manager.store_text("Utinni!", "jawa.txt")
        manager.store_bytes(b"Yub Nub!", "ewok")
        manager.store_json(
            dict(
                full_name="Jabba Desilijic Tiure",
                age=604,
                quote="beeska chata wnow kong bantha poodoo!",
            ),
            "hutt/jabba.json",
        )

    cli()
